/*
** Copyright (C) 2002 Tom Nicolai <nicolai@tzi.de>, Nils Behrens <psi@tzi.de>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/
#include <mpn/matrix.h>
#include "matrix_intern.h"
#include "md5.h"
#include <string.h>
#include <time.h>

#include <libxml/nanohttp.h>
#include <locale.h>

//only for debugging...
//#include "sflsyst.h"

//Unix only
#if (defined (__UNIX__))
#include <unistd.h>
#endif
//end Unix only


/* global things */

char workdir[MPN_STRING_LEN];
char newsfile[MPN_STRING_LEN];
char prefsfile[MPN_STRING_LEN];
char contentdir[MPN_STRING_LEN];

/* Internal Functions */

int node_has_name( xmlNodePtr cur, const char* name ) 
{
  return( (!xmlStrcmp( cur->name, (const xmlChar *) name ) ) );
}

int node_has_name_notempty( xmlDocPtr doc, xmlNodePtr cur, const char* name )
{
  if( (!xmlStrcmp( cur->name, (const xmlChar *) name ) ) ) {
    if( cur->xmlChildrenNode ) {
      return 1;
    }
  }
  return 0;
}

int node_get_name_notempty( xmlDocPtr doc, xmlNodePtr cur, const char* name, 
			    char** content  )
{
  char* c;

  if( (!xmlStrcmp( cur->name, (const xmlChar *) name ) ) ) {
    if( cur->xmlChildrenNode ) {
      c = (char*)(xmlNodeListGetString( doc, cur->xmlChildrenNode, 1));
      if( c ) {
	*content = c;
	return 1;
      }
    }
  }
  return 0;
}


void delete_node( xmlDocPtr doc, xmlNodePtr cur, const char* node )
{
  cur = cur->xmlChildrenNode;
  while( cur ) {
    if( xmlIsBlankNode( cur ) ) {
    } else if( node_has_name( cur, node ) ) {
      xmlUnlinkNode( cur );
      xmlFreeNode( cur );
      return;
    }
    cur = cur->next;
  }
}

/* External Functions */

int mpn_init()
{
  xmlDocPtr doc;
  xmlNodePtr cur;

  xmlKeepBlanksDefault(0);
  xmlNanoHTTPInit();

  sprintf( workdir, "%s/.matrix", getenv( "HOME" ) );
  sprintf( newsfile, "%s/news.xml", workdir );
  sprintf( prefsfile, "%s/prefs.xml", workdir );
  sprintf( contentdir, "%s/content", workdir );

  mfl_mkdir(workdir);
  mfl_mkdir(contentdir);

  /* Make empty Newsfile if it doesnt exist */
  doc = xmlParseFile( newsfile );
  if( doc==NULL ) {
    doc = xmlNewDoc( "1.0" );
    xmlDocSetRootElement( doc, xmlNewNode( NULL, "matrix-private-news" ) );
    xmlSaveFormatFile( newsfile, doc, 1 );
    fprintf( stderr, "Private News file doesnt exist,\n" );
    fprintf( stderr, "Created %s.\n", newsfile );
  }
  xmlFreeDoc( doc );

  /* Ckeck for prefsfile */
  doc = xmlParseFile( prefsfile );
  if (doc == NULL) {
    fprintf(stderr, "You need a %s file first!\n", prefsfile );
    return MPN_PERM_ERR;
  }
  cur = xmlDocGetRootElement(doc);
  if (cur == NULL) {
    fprintf(stderr, "File must not be empty: %s.\n", prefsfile );
    xmlFreeDoc(doc);
    return MPN_PERM_ERR;
  }
  if (xmlStrcmp(cur->name, (const xmlChar *) "matrix-private-prefs")) {
    fprintf(stderr, "File of wrong type: %s!\n", prefsfile );
    fprintf(stderr,"Root node != matrix-private-prefs.\n");
    xmlFreeDoc(doc);
    return MPN_PERM_ERR;
  }
  xmlFreeDoc( doc );

  return MPN_OK;
}

int mpn_cleanup()
{
  xmlNanoHTTPCleanup();
  xmlCleanupParser();
  return MPN_OK;
}

const char* mpn_get_workdir()
{
  return workdir;
}

const char* mpn_get_newsfile()
{
  return newsfile;
}

const char* mpn_get_prefsfile()
{
  return prefsfile;
}

const char* mpn_get_contentdir()
{
  return contentdir;
}

/* NEWS NEWS NEWS NEWS NEWS */

/* Dealing with news files */

/* Internal functions */

int lock_file( const char* filename )
{
  FILE* fp;
  
  char lockfile[MPN_STRING_LEN];

  sprintf( lockfile, "%s.lock", filename );

  //f = open(lockfile, O_CREAT | O_EXCL );
  fp=fopen(lockfile, "r");
  if( fp!=NULL ) {
    return MPN_TEMP_ERR;
  }
  else
  {
    fp=fopen(lockfile, "w");
  }

  fclose( fp );
  return MPN_OK;
}

int check_lock( const char* filename )
{
  FILE* fp;
  char lockfile[MPN_STRING_LEN];

  sprintf( lockfile, "%s.lock", filename );

  fp = fopen( lockfile, "r" );
  if( fp==NULL ) {
    return MPN_OK;
  }
 
  fclose( fp );
  return MPN_TEMP_ERR;
}

int unlock_file( const char* filename )
{
  char lockfile[MPN_STRING_LEN];

  sprintf( lockfile, "%s.lock", filename );

  unlink( lockfile );
  return MPN_OK;
}

void free_app_block( mpn_app_block_ptr ablock )
{
  free( ablock );
}

void free_file( mpn_file_ptr file )
{
  if( file->typeblock ) free_app_block( file->typeblock );
  free( file );
}

void free_sig( mpn_sig_ptr sigitem )
{
  if( sigitem->file ) free_file( sigitem->file );
  if( sigitem->author ) free( sigitem->author );
  if( sigitem->expire ) free( sigitem->expire );
  free( sigitem );
}

void free_net( mpn_net_ptr netitem )
{
  int i;

  for( i=0; i<netitem->mirrors->len; i++ ) {
    free( netitem->mirrors->list[i] );
  }
  free( netitem->mirrors->list );
  free( netitem->mirrors );
  free( netitem );
}

void free_app( mpn_app_ptr appitem )
{
  int i;

  for( i=0; i<appitem->apps->len; i++ ) {
    free_app_block( appitem->apps->list[i] );
  }
  free( appitem->apps->list );
  free( appitem->apps );
  free( appitem );
}

void free_newsitem( mpn_newsitem_ptr newsitem )
{
  if( newsitem->key ) free( newsitem->key );
  if( newsitem->sig ) free_sig( newsitem->sig );
  if( newsitem->net ) free_net( newsitem->net );
  if( newsitem->app ) free_app( newsitem->app );
  free( newsitem );
}

mpn_key_ptr parse_key( xmlDocPtr doc, xmlNodePtr cur )
{
  char* content;
  mpn_key_ptr key;

  if( node_get_name_notempty( doc, cur, "key", (char**)&content ) ) {
    key = (mpn_key_ptr)calloc( sizeof( mpn_key ), 1 );
    strcpy( key->key, content );

#if (defined (__UNIX__))
	free( content );
#endif

    return key;
  }
  return NULL;
}

mpn_file_ptr parse_file( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_file_ptr fileitem;
  char* content;

  fileitem = (mpn_file_ptr)calloc( sizeof( mpn_file ), 1 );
  
  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {

    } else if( node_get_name_notempty( doc, cur, "url", (char**)&content ) ) {
      strcpy( fileitem->url, content );

#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "hash", (char**)&content ) ) {
      strcpy( fileitem->hash, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "size", (char**)&content ) ) {
      fileitem->size = atol( content );

#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_has_name( cur, "type" ) ) {

      content = (char*)xmlGetProp( cur, "mime" );
      if( content ) {
	strcpy( fileitem->type, content );
	
#if (defined (__UNIX__))	
	free( content );
#endif

      }

      if( fileitem->typeblock ) {
	free_app_block( fileitem->typeblock );      
	fileitem->typeblock = NULL;
      }
      if( cur->xmlChildrenNode ) {
	fileitem->typeblock = calloc( sizeof(mpn_app_block), 1 );
	fileitem->typeblock->cur = cur;
      }
    }
    cur = cur->next;
  }
  
  if( fileitem->url[0] && fileitem->type[0] ) {
    return fileitem;

  } else {
    free_file( fileitem );
    return NULL;
  }
}

mpn_person_ptr parse_person( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_person_ptr personitem;
  char* content;

  personitem = (mpn_person_ptr)calloc( sizeof( mpn_person ), 1 );
  
  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {

    } else if( node_get_name_notempty( doc, cur, "name", (char**)&content ) ) {
      strcpy( personitem->name, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "home", (char**)&content ) ) {
      strcpy( personitem->home, content );

#if (defined (__UNIX__))      
	  free( content );
#endif

    }
    cur = cur->next;
  }
  
  if( personitem->home ) {
    return personitem;
  } else {
    free( personitem );
    return NULL;
  }
}

mpn_date_ptr parse_date( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_date_ptr dateitem;
  char* content;

  dateitem = (mpn_date_ptr)calloc( sizeof( mpn_date ), 1 );
  
  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {

    } else if( node_get_name_notempty( doc, cur, "sec", (char**)&content ) ) {
      dateitem->sec = atol( content );
      
#if (defined (__UNIX__))	  
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "txt", (char**)&content ) ) {
      strcpy( dateitem->txt, content );
      
#if (defined (__UNIX__))	  
	  free( content );
#endif

    }
    cur = cur->next;
  }
  
  if( dateitem->sec ) {
    return dateitem;
  } else {
    free( dateitem );
    return NULL;
  }
}

mpn_sig_ptr parse_sig( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_sig_ptr sigitem;
  char* content;
  mpn_file_ptr fileitem;
  mpn_person_ptr personitem;
  mpn_date_ptr dateitem;

  sigitem = (mpn_sig_ptr)calloc( sizeof( mpn_sig ), 1 );
  
  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {

    } else if( node_get_name_notempty( doc, cur, "id", (char**)&content ) ) {
      strcpy( sigitem->id, content );
      
#if (defined (__UNIX__))	  
	  free( content );
#endif

    } else if( node_has_name_notempty( doc, cur, "file" ) ) {
      fileitem = parse_file( doc, cur );
      if( fileitem ) {
	if( sigitem->file ) free( sigitem->file );
	sigitem->file = fileitem;
      }

    } else if( node_has_name_notempty( doc, cur, "author" ) ) {
      personitem = parse_person( doc, cur );
      if( personitem ) {
	if( sigitem->author ) free( sigitem->author );
	sigitem->author = personitem;
      }

    } else if( node_has_name_notempty( doc, cur, "expire" ) ) {
      dateitem = parse_date( doc, cur );
      if( dateitem ) {
	if( sigitem->expire ) free( sigitem->expire );
	sigitem->expire = dateitem;
      }

    }
    cur = cur->next;
  }
  
  if( sigitem->id[0] ) {
    return sigitem;
  } else {
    free_sig( sigitem );
    return NULL;
  }
}

mpn_mirror_ptr parse_mirror( xmlDocPtr doc, xmlNodePtr cur )
{
  char* content;
  mpn_mirror_ptr mirror;

  if( node_get_name_notempty( doc, cur, "mirror", (char**)&content ) ) {
    mirror = (mpn_mirror_ptr)calloc( sizeof( mpn_mirror ), 1 );
    strcpy( mirror->url, content );
    
#if (defined (__UNIX__))	
	free( content );
#endif

    return mirror;
  }
  return NULL;
}

mpn_net_ptr parse_net( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_net_ptr netitem;
  char* content;
  mpn_mirror_ptr mirroritem;

  netitem = (mpn_net_ptr)calloc( sizeof( mpn_net ), 1 );
  netitem->cur = cur;
  netitem->mirrors = (mpn_mirror_list_ptr)calloc( sizeof( mpn_mirror_list ),1);

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {

    } else if( node_get_name_notempty( doc, cur,"weight",(char**)&content ) ) {
      netitem->weight = atof( content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur,"ttl",(char**)&content ) ) {
      netitem->ttl = atoi( content );
      
#if (defined (__UNIX__))	  
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur,"hop",(char**)&content ) ) {
      netitem->hop = atoi( content );

#if (defined (__UNIX__))
      free( content );
#endif

    } else if( node_has_name( cur, "corrupt" ) ) {
      netitem->corrupt = 1;

    } else if( node_has_name( cur, "keep" ) ) {
      netitem->keep = 1;

    } else if( node_has_name( cur, "dont_recommend" ) ) {
      netitem->dontrecommend = 1;

    } else if( node_get_name_notempty( doc, cur,"localfile",(char**)&content ) ) {
      strcpy( netitem->localfile, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_has_name_notempty( doc, cur, "mirror" ) ) {
      mirroritem = parse_mirror( doc, cur );
      if( mirroritem ) {
	netitem->mirrors->list = (mpn_mirror_ptr*)realloc( 
                     netitem->mirrors->list, 
		     sizeof( mpn_mirror_ptr ) * (netitem->mirrors->len+1) );
	netitem->mirrors->list[netitem->mirrors->len] = mirroritem;
	netitem->mirrors->len++;
      }

    }
    cur = cur->next;
  }

  return netitem;
}

mpn_app_block_ptr parse_appblock( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_app_block_ptr blockitem;

  blockitem = (mpn_app_block_ptr)calloc( sizeof( mpn_app_block ), 1 );
  strcpy( blockitem->name, (const char*)cur->name );
  blockitem->cur = cur;
  
  return blockitem;
}

mpn_app_ptr parse_app( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_app_ptr appitem;
  mpn_app_block_ptr blockitem;

  appitem = (mpn_app_ptr)calloc( sizeof( mpn_app ), 1 );
  appitem->cur = cur;
  appitem->apps = (mpn_app_list_ptr)calloc( sizeof( mpn_app_list ), 1 );

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else {
      blockitem = parse_appblock( doc, cur );
      if( blockitem ) {
	appitem->apps->list = (mpn_app_block_ptr*)realloc( 
                     appitem->apps->list, 
		     sizeof( mpn_app_block_ptr ) * (appitem->apps->len+1) );
	appitem->apps->list[appitem->apps->len] = blockitem;
	appitem->apps->len++;
      }
    }
    cur = cur->next;
  }

  return appitem;
}

mpn_newsitem_ptr parse_news( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_newsitem_ptr newsitem;
  mpn_key_ptr keyitem;
  mpn_sig_ptr sigitem;
  mpn_net_ptr netitem;
  mpn_app_ptr appitem;

  newsitem = (mpn_newsitem_ptr)calloc( sizeof( mpn_newsitem ), 1 );
  newsitem->cur = cur;

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {

    } else if( node_has_name_notempty( doc, cur, "key" ) ) {
      keyitem = parse_key( doc, cur );
      if( keyitem ) {
	if( newsitem->key ) free( newsitem->key );
	newsitem->key = keyitem;
      }

    } else if( node_has_name_notempty( doc, cur, "sig" ) ) {
      sigitem = parse_sig( doc, cur );
      if( sigitem ) {
	if( newsitem->sig ) free_sig( newsitem->sig );
	newsitem->sig = sigitem;
      }

    } else if( node_has_name_notempty( doc, cur, "net" ) ) {
      netitem = parse_net( doc, cur );
      if( netitem ) {
	if( newsitem->net ) free_net( newsitem->net );
	newsitem->net = netitem;
      }

    } else if( node_has_name_notempty( doc, cur, "app" ) ) {
      appitem = parse_app( doc, cur );
      if( appitem ) {
	if( newsitem->app ) free_app( newsitem->app );
	newsitem->app = appitem;
      }

    }
    cur = cur->next;
  }  

  if( newsitem->sig == NULL ) {
    free_newsitem( newsitem );
    return NULL;

  } else {
    return newsitem;
  }
}

int parse_news_doc( xmlDocPtr doc, mpn_news_ptr* newsdoc_ret )
{
  xmlNodePtr cur;
  char* content;
  mpn_newsitem_ptr newsitem;
  mpn_news_ptr newsdoc;
  int type=-1;

  cur = xmlDocGetRootElement( doc );
  if (cur == NULL) {
    fprintf(stderr,"Error: empty document.\n" );
    return MPN_PERM_ERR;
  }
  if (xmlStrcmp(cur->name, (const xmlChar *)"matrix-public-news" )==0) {
    type = MPN_NEWS_PUBLIC;
  } else if (xmlStrcmp(cur->name, (const xmlChar *)"matrix-private-news")==0) {
    type = MPN_NEWS_PRIVATE;
  } else if (xmlStrcmp(cur->name, (const xmlChar *)"matrix-publish" )==0) {
    type = MPN_NEWS_PUBLISH;
  } else if (xmlStrcmp(cur->name, (const xmlChar *)"matrix-recommend" )==0) {
    type = MPN_NEWS_RECOMMEND;
  } else if (xmlStrcmp(cur->name, (const xmlChar *)"matrix-temp-news" )==0) {
    type = MPN_NEWS_TEMP;
  } else {
    //fprintf(stderr,"Error: document of unknown type\n" );
    type = MPN_NEWS_UNKNOWN;
    //return MPN_PERM_ERR;
  }

  /* Build mpn_news */
  newsdoc = calloc( sizeof( mpn_news ), 1 );
  newsdoc->mode = MPN_OPEN_READONLY;
  newsdoc->doc.doc = doc;
  newsdoc->cur = cur;
  newsdoc->news_type = type;
  strcpy( newsdoc->root, (const char*)cur->name );

  newsdoc->news = calloc( sizeof( mpn_newslist ), 1 );

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if(node_get_name_notempty(doc,cur,"cookie",(char**)&content)){
      strcpy( newsdoc->cookie, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if(node_get_name_notempty(doc,cur,"username",(char**)&content)){
      strcpy( newsdoc->username, content );
      
#if (defined (__UNIX__))	  
	  free( content );
#endif

    } else if(node_has_name(cur,"mailreply")){
      newsdoc->mailreply=1;

    } else if( node_has_name_notempty( doc, cur, "news" ) ) {
      newsitem = parse_news( doc, cur );
      if( newsitem ) {
	newsdoc->news->list = (mpn_newsitem_ptr*)realloc( newsdoc->news->list, 
		     sizeof( mpn_newsitem_ptr ) * (newsdoc->news->len+1) );
	newsdoc->news->list[newsdoc->news->len] = newsitem;
	newsdoc->news->len++;
      }
    }

    cur = cur->next;
  }

  *newsdoc_ret = newsdoc;
  return MPN_OK;
}

int mpn_news_open_file_once( const char* filename, int mode, 
			     mpn_news_ptr* newsdoc_ret, const char* root )
{
  int r;
  mpn_news_ptr newsdoc;
  xmlDocPtr doc;

  /* Do Lock Stuff */
  if( mode==MPN_OPEN_READONLY ) {
    r = check_lock( filename );
    if( r!=MPN_OK ) return r;

  } else if( mode==MPN_OPEN_READWRITE ) {
    r = lock_file( filename );
    if( r!=MPN_OK ) return r;

  } else {
    return MPN_PERM_ERR;
  }

  /* Open and Read */
  doc = xmlParseFile( filename );
  if( doc==NULL ) {
    doc = xmlNewDoc( "1.0" );
    xmlDocSetRootElement( doc, xmlNewNode( NULL, root ) );
    xmlSaveFormatFile( filename, doc, 1 );
    xmlFreeDoc( doc );

    doc = xmlParseFile( filename );
    if( doc==NULL ) {

      fprintf( stderr, "Error: Newsfile %s doesnt exist and cannot be created.\n", filename );
      if( mode==MPN_OPEN_READWRITE ) unlock_file( filename );
      return MPN_PERM_ERR;
    }
  }
  r = parse_news_doc( doc, &newsdoc );

  /* Everything good? */
  if( r!=MPN_OK ) {
    xmlFreeDoc( doc );
    if( mode==MPN_OPEN_READWRITE ) unlock_file( filename );
    return r;

  } else {
    newsdoc->mode = mode;
    //strcpy( newsdoc->root, root );
    *newsdoc_ret = newsdoc;
    return MPN_OK;
  }
}

int mpn_news_open_file_retry( const char* filename, int mode, 
			      mpn_news_ptr* newsdoc_ret, const char* root )
{
  int i, r;
  mpn_news_ptr newsdoc;

  for( i=0, r=MPN_TEMP_ERR; (i<10)&&(r!=MPN_OK); i++ ) {
    r = mpn_news_open_file_once( filename, mode, &newsdoc, root );
    if( r==MPN_TEMP_ERR ) {
      
      mfl_sleep(500);
      printf( "open news retry\n" );
      if( i==9 ) printf( "Error: cannot open file %s.\n  Either your computer is under heavy load,\n  or a crashed program left the %s.lock file.\n  If this problem returns delete the .lock file.\n", filename, filename );
    }
  }

  if( r==MPN_OK ) {
    strcpy( newsdoc->filename, filename );
    *newsdoc_ret = newsdoc;
    return MPN_OK;
  } else {
    return MPN_PERM_ERR;
  }
}

/* External functions */

void mpn_news_open_empty( mpn_news_ptr* newsdoc_ret )
{
  mpn_news_ptr news;

  news = calloc( sizeof( mpn_news ), 1 );
  news->mode = MPN_OPEN_READONLY;
  //strcpy( news->root, "matrix-temp-news" );
  news->news = calloc( sizeof( mpn_newslist ), 1 );
  
  news->doc.doc = xmlNewDoc( "1.0" );
  news->cur = xmlNewNode( NULL, "matrix-temp-news" );
  xmlDocSetRootElement( news->doc.doc, news->cur );
  
  *newsdoc_ret = news;
}

int mpn_news_open_private( int mode, mpn_news_ptr* newsdoc_ret )
{
  return mpn_news_open_file_retry( mpn_get_newsfile(), mode, newsdoc_ret,
				   "matrix-private-news" );
}

int mpn_news_open_foreign( const char* filename, int mode, 
			   mpn_news_ptr* newsdoc_ret, const char* root )
{
  return mpn_news_open_file_retry( filename, mode, newsdoc_ret, root );
}

int mpn_news_open_mem( const char* buf, mpn_news_ptr* newsdoc_ret )
{
  int r;
  mpn_news_ptr newsdoc;
  xmlDocPtr doc;

  /* Open and Read */
  doc = xmlParseMemory( buf, strlen( buf ) );
  if( doc==NULL ) {
    fprintf( stderr, "Error: Newsfile doesnt exist.\n" );
    return MPN_PERM_ERR;
  }
  r = parse_news_doc( doc, &newsdoc );

  /* Everything good? */
  if( r!=MPN_OK ) {
    xmlFreeDoc( doc );
    return r;

  } else {
    newsdoc->mode = MPN_OPEN_READONLY;
    *newsdoc_ret = newsdoc;
    return MPN_OK;
  }
}

void mpn_news_close( mpn_news_ptr newsdoc )
{
  int i;

  if( newsdoc->mode==MPN_OPEN_READWRITE ) {
    xmlSaveFormatFile( newsdoc->filename, newsdoc->doc.doc, 1 );
    unlock_file( newsdoc->filename );
  }

  xmlFreeDoc( newsdoc->doc.doc );

  if( newsdoc->news->len>0 ) {
    for( i=0; i<newsdoc->news->len; i++ ) {
      free_newsitem( newsdoc->news->list[i] );
    }
    free( newsdoc->news->list );
  }
  free( newsdoc->news );
  free( newsdoc );
}

int mpn_fetch_newslist( const char* url, const char* cookie, 
			const char* filename )
{
  char buf[MPN_STRING_LEN];
  int r;

  strcpy( buf, url );

  if( strchr( buf, '?' ) ) strcat( buf, "&" );
  else strcat( buf, "?" );

  strcat( buf, "cookie=" );
  strcat( buf, cookie );

  r = xmlNanoHTTPFetch( buf, filename, NULL );
  if( r==0 ) return MPN_OK;
  else return MPN_PERM_ERR;
}

/* Dealing with filter docs */

int mpn_xslfilterdoc_open( const char* filename,
			   mpn_xslfilterdoc_ptr* filter_ret )
{
  char buf[MPN_STRING_LEN];
  mpn_xslfilterdoc_ptr filterdoc;

  sprintf( buf, "%s/%s", mpn_get_workdir(), filename );

  filterdoc = calloc( sizeof( mpn_xslfilterdoc ), 1 );

  filterdoc->doc = xmlParseFile( buf );
  if( filterdoc->doc==NULL ) {
    fprintf( stderr, "Error: Filterfile %s doesnt exist.\n", buf );
    free( filterdoc );
    return MPN_PERM_ERR;
  }

  filterdoc->xsl = xsltParseStylesheetDoc( filterdoc->doc );
  if( filterdoc->xsl==NULL ) {
    fprintf( stderr, "Error: Filterfile %s is no XSL File.\n", buf );
    xmlFreeDoc( filterdoc->doc );
    free( filterdoc );
    return MPN_PERM_ERR;
  }

  *filter_ret = filterdoc;
  return MPN_OK;
}

void mpn_xslfilterdoc_close( mpn_xslfilterdoc_ptr filterdoc )
{
  xsltFreeStylesheet( filterdoc->xsl );
  free( filterdoc );
}

/* Dealing with news */

mpn_docinfo_ptr mpn_news_get_docinfo( mpn_news_ptr newsdoc )
{
  return &newsdoc->doc;
}

int mpn_news_get_type( mpn_news_ptr newsdoc )
{
  return newsdoc->news_type;
}

const char* mpn_news_get_rootnode( mpn_news_ptr newsdoc )
{
  return newsdoc->root;
}

const char* mpn_news_get_username( mpn_news_ptr newsdoc )
{
  return newsdoc->username;
}

int mpn_news_get_mailreply( mpn_news_ptr newsdoc )
{
  return newsdoc->mailreply;
}

const char* mpn_news_get_cookie( mpn_news_ptr newsdoc )
{
  return newsdoc->cookie;
}

const char* mpn_news_get_filename( mpn_news_ptr newsdoc )
{
  return newsdoc->filename;
}

int mpn_news_get_mode( mpn_news_ptr newsdoc )
{
  return newsdoc->mode;
}

int mpn_news_get_len( mpn_news_ptr newsdoc )
{
  return newsdoc->news->len;
}

mpn_newsitem_ptr mpn_news_get_item( mpn_news_ptr newsdoc, int i )
{
  return newsdoc->news->list[i];
}

void mpn_news_add_item( mpn_news_ptr newsdoc, mpn_newsitem_ptr item )
{
  mpn_newsitem_ptr newitem;
  xmlNodePtr cur;

  cur = xmlAddChild( newsdoc->cur, xmlCopyNode( item->cur, 1 ) );
  newitem = parse_news( newsdoc->doc.doc, cur );

  newsdoc->news->list = (mpn_newsitem_ptr*)realloc( newsdoc->news->list, 
		     sizeof( mpn_newsitem_ptr ) * (newsdoc->news->len+1) );
  newsdoc->news->list[newsdoc->news->len] = newitem;
  newsdoc->news->len++;

}

void mpn_news_del_item( mpn_news_ptr newsdoc, mpn_newsitem_ptr item )
{
  int i, j;

  xmlUnlinkNode( item->cur );
  xmlFreeNode( item->cur );

  for( j=-1, i=0; i<newsdoc->news->len; i++ ) {
    if( item==newsdoc->news->list[i] ) j=i;
  }

  if( j==-1 ) {
    fprintf( stderr, "mpn_news_del_item: item not in newsdoc!\n" );
    return;
  }
    
  for( i=j; i<newsdoc->news->len-1; i++ ) {
    newsdoc->news->list[i] = newsdoc->news->list[i+1];
  }

  free_newsitem( item );
  newsdoc->news->len--;
  newsdoc->news->list = (mpn_newsitem_ptr*)realloc( newsdoc->news->list, 
		   sizeof( mpn_newsitem_ptr ) * newsdoc->news->len );
}

int mpn_news_filter( mpn_news_ptr newsdoc, mpn_filter_ptr filter,
		     mpn_news_ptr* newsdoc_ret, int verbose )
{
  mpn_news_ptr ndoc;
  mpn_news_ptr fdoc;
  int r, i;

  mpn_news_open_empty( &ndoc );

  for( i=0; i<mpn_news_get_len( newsdoc ); i++ ) {
    r = mpn_newsitem_filter( mpn_news_get_item( newsdoc, i ),
			     filter, &fdoc, verbose );
    if( r==MPN_OK ) {
      mpn_news_add_newsdoc( ndoc, fdoc );
      mpn_news_close( fdoc );
    } else {
      mpn_news_close( ndoc );
      return r;
    }
  }

  *newsdoc_ret = ndoc;
  return MPN_OK;  
}

void mpn_news_add_newsdoc( mpn_news_ptr newsdoc, mpn_news_ptr adddoc )
{
  int i;

  for( i=0; i<mpn_news_get_len( adddoc ); i++ ) {
    mpn_news_add_item( newsdoc, mpn_news_get_item( adddoc, i ) );
  }
}

int mpn_news_compose_default( mpn_identity_ptr identity,
			      mpn_news_ptr* newsdoc_ret,
			      const char* idpart, const char* idpart2,
			      long expire, 
			      const char* subject, const char* location )
{ 
  mpn_news_ptr newsdoc;
  mpn_person_ptr person;
  int r;
  char composition[5000];
  long timestamp;
  char id[MPN_STRING_LEN];
  char timestamp_str[MPN_STRING_LEN];
  char expire_str[MPN_STRING_LEN];
  xmlDocPtr doc;

  person = mpn_identity_get_person( identity );

  // Make Timestamp
  timestamp = time( NULL );
  sprintf( id, "%s-%ld%s", idpart, timestamp, idpart2 );
  strcpy( timestamp_str, ctime( &timestamp ) );
  strcpy( expire_str, ctime( &expire ) );

  sprintf( composition, "\
<?xml version=\"1.0\"?>\
<matrix-temp-news>\
  <news version=\"1\">\
    <sig>\
      <id>%s</id>\
      <author>\
        <name>%s</name>\
        <home>%s</home>\
      </author>\
      <subject>%s</subject>\
      <location>%s</location>\
      <publish_time>\
        <sec>%ld</sec>\
        <txt>%s</txt>\
      </publish_time>\
      <expire>\
        <sec>%ld</sec>\
        <txt>%s</txt>\
      </expire>\
    </sig>\
    <net>\
      <ttl>%d</ttl>\
      <hop>0</hop>\
      <weight>1</weight>\
    </net>\
    <app>\
      <heal_the_world>format c:</heal_the_world>\
    </app>\
  </news>\
</matrix-temp-news>", 
	   id, 
	   mpn_person_get_name( person ),
	   mpn_person_get_home( person ),
	   subject, location,
	   timestamp, timestamp_str,
	   expire, expire_str, 5 );

  doc = xmlParseMemory( composition, strlen( composition ) );
  r = parse_news_doc( doc, &newsdoc );

  if( r!=MPN_OK ) {
    xmlFreeDoc( doc );
    return r;

  } else {
    *newsdoc_ret = newsdoc;
    return MPN_OK;
  }
}

int mpn_news_compose_default_file( mpn_identity_ptr identity,
				   mpn_news_ptr* newsdoc_ret,
				   const char* idpart, const char* idpart2,
				   const char* filebuf, 
				   long filesize, const char* type, 
				   const char* typecont, long expire,
				   const char* subject, const char* location )
{ 
  mpn_news_ptr newsdoc;
  mpn_person_ptr person;
  int r;
  char composition[5000];
  long timestamp;
  char md5hash[20];
  char strhash[MPN_STRING_LEN];
  char id[MPN_STRING_LEN];
  char url[MPN_STRING_LEN];
  char timestamp_str[MPN_STRING_LEN];
  char expire_str[MPN_STRING_LEN];
  xmlDocPtr doc;
  
  person = mpn_identity_get_person( identity );

  // Make Timestamp
  timestamp = time( NULL );
  strcpy( timestamp_str, ctime( &timestamp ) );
  strcpy( expire_str, ctime( &expire ) );

  // Calc md5 string
  md5_buffer( filebuf, filesize, md5hash );
  md5_sig_to_string( md5hash, strhash, MPN_STRING_LEN );
  sprintf( id, "%s-%ld-%s%s", idpart, timestamp, strhash, idpart2 );
  
  //  sprintf( id, "%s-ld-%d", idpart, timestamp, rand() );

  // Make URL
  sprintf( url, "%s/%s", mpn_identity_get_public_space( identity ), id );

  sprintf( composition, "\
<?xml version=\"1.0\"?>\
<matrix-temp-news>\
  <news version=\"1\">\
    <sig>\
      <id>%s</id>\
      <file>\
        <url>%s</url>\
        <hash>%s</hash>\
        <size>%ld</size>\
        <type mime=\"%s\">\
          %s\
        </type>\
      </file>\
      <author>\
        <name>%s</name>\
        <home>%s</home>\
      </author>\
      <subject>%s</subject>\
      <location>%s</location>\
      <publish_time>\
        <sec>%ld</sec>\
        <txt>%s</txt>\
      </publish_time>\
      <expire>\
        <sec>%ld</sec>\
        <txt>%s</txt>\
      </expire>\
    </sig>\
    <net>\
      <ttl>%d</ttl>\
      <hop>0</hop>\
      <weight>1</weight>\
    </net>\
    <app>\
      <heal_the_world>format c:</heal_the_world>\
    </app>\
  </news>\
</matrix-temp-news>", 
	   id, url, strhash, filesize, type, typecont,
	   mpn_person_get_name( person ),
	   mpn_person_get_home( person ),
	   subject, location,
	   timestamp, timestamp_str,
	   expire, expire_str, 5 );

  doc = xmlParseMemory( composition, strlen( composition ) );
  r = parse_news_doc( doc, &newsdoc );

  if( r!=MPN_OK ) {
    xmlFreeDoc( doc );
    return r;

  } else {
    *newsdoc_ret = newsdoc;
    return MPN_OK;
  }
}

int mpn_news_compose_default_url( mpn_identity_ptr identity,
				  mpn_news_ptr* newsdoc_ret,
				  const char* idpart, const char* idpart2,
				  const char* url, 
				  const char* type, const char* typecont, 
				  long expire,
				  const char* subject, const char* location )
{ 
  mpn_news_ptr newsdoc;
  mpn_person_ptr person;
  int r;
  char composition[5000];
  long timestamp;
  char id[MPN_STRING_LEN];
  char timestamp_str[MPN_STRING_LEN];
  char expire_str[MPN_STRING_LEN];
  xmlDocPtr doc;
  
  person = mpn_identity_get_person( identity );

  // Make Timestamp
  timestamp = time( NULL );
  strcpy( timestamp_str, ctime( &timestamp ) );
  strcpy( expire_str, ctime( &expire ) );

  // Make ID
  sprintf( id, "%s-%ld-%d%s", idpart, timestamp, rand(), idpart2 );

  sprintf( composition, "\
<?xml version=\"1.0\"?>\
<matrix-temp-news>\
  <news version=\"1\">\
    <sig>\
      <id>%s</id>\
      <file>\
        <url>%s</url>\
        <type mime=\"%s\">\
          %s\
        </type>\
      </file>\
      <author>\
        <name>%s</name>\
        <home>%s</home>\
      </author>\
      <subject>%s</subject>\
      <location>%s</location>\
      <publish_time>\
        <sec>%ld</sec>\
        <txt>%s</txt>\
      </publish_time>\
      <expire>\
        <sec>%ld</sec>\
        <txt>%s</txt>\
      </expire>\
    </sig>\
    <net>\
      <ttl>%d</ttl>\
      <hop>0</hop>\
      <weight>1</weight>\
    </net>\
    <app>\
      <heal_the_world>format c:</heal_the_world>\
    </app>\
  </news>\
</matrix-temp-news>", 
	   id, url, type, typecont,
	   mpn_person_get_name( person ),
	   mpn_person_get_home( person ),
	   subject, location,
	   timestamp, timestamp_str,
	   expire, expire_str, 5 );

  doc = xmlParseMemory( composition, strlen( composition ) );
  r = parse_news_doc( doc, &newsdoc );

  if( r!=MPN_OK ) {
    xmlFreeDoc( doc );
    return r;

  } else {
    *newsdoc_ret = newsdoc;
    return MPN_OK;
  }
}

/* Dealing with newsitems */

mpn_key_ptr mpn_newsitem_get_key( mpn_newsitem_ptr item )
{
  return item->key;
}

mpn_sig_ptr mpn_newsitem_get_sig( mpn_newsitem_ptr item )
{
  return item->sig;
}

mpn_net_ptr mpn_newsitem_get_net( mpn_newsitem_ptr item )
{
  return item->net;
}

mpn_app_ptr mpn_newsitem_get_app( mpn_newsitem_ptr item )
{
  return item->app;
}

int mpn_newsitem_equal( mpn_newsitem_ptr i1, mpn_newsitem_ptr i2 )
{
  mpn_sig_ptr sig1, sig2;

  sig1 = mpn_newsitem_get_sig( i1 );
  if( !sig1 ) return 0;
  sig2 = mpn_newsitem_get_sig( i2 );
  if( !sig2 ) return 0;

  return ( strcmp( mpn_sig_get_id( sig1 ), mpn_sig_get_id( sig2 ) )==0 );
}

int mpn_newsitem_filter( mpn_newsitem_ptr item, mpn_filter_ptr filter,
			 mpn_news_ptr* newsdoc_ret, int verbose )
{
  mpn_xslfilterdoc_ptr xsldoc;
  int r;

  r = mpn_xslfilterdoc_open( mpn_filter_get_xslprocess( filter ), &xsldoc );
  if( r!=MPN_OK ) {
    fprintf( stderr, "error opening global filter\n" );
    return r;
  }
  
  r = mpn_newsitem_xsl_process( item, xsldoc, newsdoc_ret, verbose );
  
  mpn_xslfilterdoc_close( xsldoc );
  return r;
}

int mpn_newsitem_xsl_process( mpn_newsitem_ptr item,
			      mpn_xslfilterdoc_ptr xsldoc,
			      mpn_news_ptr* newsdoc_ret, int verbose )
{
  xmlDocPtr doc, resdoc;
  xmlNodePtr root, cur;
  const char* params[1];
  int r;
  mpn_news_ptr newsdoc;
  xmlChar* buffer;
  int size;

  params[0] = NULL;

  doc = xmlNewDoc( "1.0" );
  root = xmlNewNode( NULL, "matrix-temp-news" );
  xmlDocSetRootElement( doc, root );
  cur = xmlAddChild( root, xmlCopyNode( item->cur, 1 ) );

  resdoc = xsltApplyStylesheet( xsldoc->xsl, doc, params );
  xmlFreeDoc( doc );

  if( verbose ) {
    xmlDocDumpFormatMemory( resdoc, &buffer, &size, 1 );
    printf( "\nMessage after filtering:\n%s\n", buffer );
    free( buffer );
  }
  
  r = parse_news_doc( resdoc, &newsdoc );

  if( r!=MPN_OK ) {
    xmlFreeDoc( resdoc );
    return r;

  } else {
    *newsdoc_ret = newsdoc;
    return MPN_OK;
  }
}

void mpn_newsitem_dump( mpn_news_ptr newsdoc, mpn_newsitem_ptr item,
			char* buf )
{
  xmlBufferPtr xmlbuffer;

  // Dump newsitem
  xmlbuffer = xmlBufferCreate();
  xmlNodeDump( xmlbuffer, newsdoc->doc.doc, item->cur, 1, 1 );

  strcpy( buf, xmlBufferContent( xmlbuffer ) );
  xmlBufferFree( xmlbuffer );
}

int mpn_newsitem_fetch_file( mpn_news_ptr newsdoc, mpn_newsitem_ptr item, 
			     char* filename_ret )
{
  mpn_news_ptr newswrite;
  mpn_newsitem_ptr itemwrite;
  mpn_sig_ptr sig;
  mpn_net_ptr netwrite;
  mpn_file_ptr file;
  char filename[MPN_STRING_LEN];
  const char* id;
  const char* url;
  const char* newsfilename;
  int r, success, closeagain, i;

#if (defined (__WINDOWS__))
  int buflen=4000;
  char filebuffer[4000];
  FILE *fp;
  //char **contentType;
  char contentType[200];
  void *ctx;
  int read;
  int write;
#endif
  
  sig = mpn_newsitem_get_sig( item );
  file = mpn_sig_get_file( sig );
  if( !file ) return MPN_PERM_ERR;

  id = mpn_sig_get_id( sig );
  if( strchr( id, '/' ) ) {
    fprintf( stderr, "malicious '/' in id, not fetching message\n" );
    return MPN_PERM_ERR;
  }

  sprintf( filename, "%s/%s", mpn_get_contentdir(), id );
  unlink( filename );
  
  url = mpn_file_get_url( file );
  if( !url ) return MPN_PERM_ERR;


//#if (defined (__UNIX__))
  r = xmlNanoHTTPFetch( url, filename, NULL );
//#endif

/*#if (defined (__WINDOWS__))
  fp=fopen(filename, "wb");
  
  ctx=xmlNanoHTTPOpen(url, &contentType);
  read=1;
  while(read!=0)
  {
	read=xmlNanoHTTPRead(ctx, filebuffer, buflen);
  			
	if(read>0) write=fwrite(filebuffer, 1, read, fp);
  }
  xmlNanoHTTPClose(ctx);

  r=fclose(fp);
#endif*/
  
  
  if( r!=0 ) {
    // try the mirrors
    success = 0;
  } else {
    success = 1;
  }

  // Update Newsdoc
  newsfilename = mpn_news_get_filename( newsdoc );

  if( newsfilename ) {

    if( mpn_news_get_mode( newsdoc ) == MPN_OPEN_READONLY ) {
      // open readwrite
      r = mpn_news_open_foreign(newsfilename,MPN_OPEN_READWRITE,&newswrite,
				mpn_news_get_rootnode(newsdoc) );
      closeagain = 1;
    } else {
      newswrite = newsdoc;
      closeagain = 0;
      r = MPN_OK;
    }

    if( r==MPN_OK ) {
      
      for( i=0; i<mpn_news_get_len( newswrite ); i++ ) {

	itemwrite = mpn_news_get_item( newswrite, i );

	if( mpn_newsitem_equal( itemwrite, item ) ) {
	  netwrite = mpn_newsitem_get_net( itemwrite );

	  if( success ) {
	    mpn_net_set_localfile( newswrite, netwrite, id );
	  } else {
	    mpn_net_set_corrupt( newswrite, netwrite, 1 );
	  }
	  break;
	}

      }

      if( closeagain ) {
	mpn_news_close( newswrite );
      }
    }

  }

  if( success ) {
    strcpy( filename_ret, id );
    return MPN_OK;
  } else {
    return MPN_PERM_ERR;
  }
}

/* Dealing with keys */

const char* mpn_key_get_signature( mpn_key_ptr key )
{
  return key->key;
}

/* Dealing with sigs */

const char* mpn_sig_get_id( mpn_sig_ptr sig )
{
  return sig->id;
}

mpn_file_ptr mpn_sig_get_file( mpn_sig_ptr sig )
{
  return sig->file;
}

mpn_person_ptr mpn_sig_get_author( mpn_sig_ptr sig )
{
  return sig->author;
}

mpn_date_ptr mpn_sig_get_expire( mpn_sig_ptr sig )
{
  return sig->expire;
}

/* Dealing with files */

const char* mpn_file_get_url( mpn_file_ptr file )
{
  return file->url;
}

const char* mpn_file_get_hash( mpn_file_ptr file )
{
  return file->hash;
}

long mpn_file_get_size( mpn_file_ptr file )
{
  return file->size;
}

const char* mpn_file_get_type( mpn_file_ptr file )
{
  return file->type;
}

mpn_app_block_ptr mpn_file_get_typeblock( mpn_file_ptr file )
{
  return file->typeblock;
}

/* Dealing with persons */

const char* mpn_person_get_name( mpn_person_ptr person )
{
  return person->name;
}

const char* mpn_person_get_home( mpn_person_ptr person )
{
  return person->home;
}

/* Dealing with dates */

long mpn_date_get_sec( mpn_date_ptr date )
{
  return date->sec;
}

const char* mpn_date_get_txt( mpn_date_ptr date )
{
  return date->txt;
}

/* Dealing withs nets */

double mpn_net_get_weight( mpn_net_ptr net )
{
  return net->weight;
}

int mpn_net_get_ttl( mpn_net_ptr net )
{
  return net->ttl;
}

int mpn_net_get_hop( mpn_net_ptr net )
{
  return net->hop;
}

int mpn_net_get_dontrecommend( mpn_net_ptr net )
{
  return net->dontrecommend;
}

int mpn_net_get_corrupt( mpn_net_ptr net )
{
  return net->corrupt;
}

int mpn_net_get_keep( mpn_net_ptr net )
{
  return net->keep;
}

const char* mpn_net_get_localfile( mpn_net_ptr net )
{
  return net->localfile;
}

int mpn_net_get_mirror_count( mpn_net_ptr net )
{
  return net->mirrors->len;
}

const char* mpn_net_get_mirror( mpn_net_ptr net, int i )
{
  return net->mirrors->list[i]->url;
}

void mpn_net_set_weight( mpn_news_ptr newsdoc, mpn_net_ptr net, 
			 double weight )
{
  char buf[MPN_STRING_LEN];
  char* oldloc;

  net->weight = weight;

  delete_node( newsdoc->doc.doc, net->cur, "weight" );

  oldloc = setlocale( LC_NUMERIC, "POSIX" );
  sprintf( buf, "%.3f", net->weight );
  setlocale( LC_NUMERIC, oldloc );

  xmlNewChild( net->cur, NULL, "weight", buf );
}

void mpn_net_set_ttl( mpn_news_ptr newsdoc, mpn_net_ptr net, 
		      int ttl )
{
  char buf[MPN_STRING_LEN];

  net->ttl = ttl;

  delete_node( newsdoc->doc.doc, net->cur, "ttl" );
  sprintf( buf, "%d", net->ttl );
  xmlNewChild( net->cur, NULL, "ttl", buf );
}

void mpn_net_set_hop( mpn_news_ptr newsdoc, mpn_net_ptr net, 
		      int hop )
{
  char buf[MPN_STRING_LEN];

  net->hop = hop;

  delete_node( newsdoc->doc.doc, net->cur, "hop" );
  sprintf( buf, "%d", net->hop );
  xmlNewChild( net->cur, NULL, "hop", buf );
}

void mpn_net_set_corrupt( mpn_news_ptr newsdoc, mpn_net_ptr net, 
			  int corrupt )
{
  net->corrupt = corrupt;

  delete_node( newsdoc->doc.doc, net->cur, "corrupt" );
  if( net->corrupt ) 
    xmlNewChild( net->cur, NULL, "corrupt", NULL );
}

void mpn_net_set_localfile( mpn_news_ptr newsdoc, mpn_net_ptr net, 
			    const char* localfile )
{
  strcpy( net->localfile, localfile );

  delete_node( newsdoc->doc.doc, net->cur, "localfile" );
  if( net->localfile[0] )
    xmlNewChild( net->cur, NULL, "localfile", net->localfile );
}

void mpn_net_add_mirror( mpn_news_ptr newsdoc, mpn_net_ptr net, 
			 const char* mirror )
{
  mpn_mirror_ptr mirroritem;
  mirroritem = calloc( sizeof( mpn_mirror ), 1 );
  strcpy( mirroritem->url, mirror );

  net->mirrors->list = (mpn_mirror_ptr*)realloc( 
		     net->mirrors->list, 
		     sizeof( mpn_mirror_ptr ) * (net->mirrors->len+1) );

  net->mirrors->list[net->mirrors->len] = mirroritem;
  net->mirrors->len++;

  xmlNewChild( net->cur, NULL, "mirror", mirror );
}

/* Dealing with apps */

mpn_app_block_ptr mpn_app_get_block( mpn_app_ptr app, const char* name )
{
  int i;
  
  for( i=0; i<app->apps->len; i++ ) {
    if( strcmp( name, app->apps->list[i]->name )==0 ) {
      return app->apps->list[i];
    }
  }
  return NULL;
}

mpn_app_block_ptr mpn_app_add_block( mpn_app_ptr app, const char* name )
{
  mpn_app_block_ptr blockitem;

  blockitem = calloc( sizeof( mpn_app_block ), 1 );
  strcpy( blockitem->name, name );
  blockitem->cur = xmlNewChild( app->cur, NULL, name, NULL );

  app->apps->list = (mpn_app_block_ptr*)realloc( 
			app->apps->list, 
			sizeof( mpn_app_block_ptr ) * (app->apps->len+1) );
  app->apps->list[app->apps->len] = blockitem;
  app->apps->len++;

  return blockitem;
}

/* Dealing with appblocks */

int mpn_app_block_get_int( mpn_docinfo_ptr docinfo, mpn_app_block_ptr ablock, 
			   const char* name )
{
  xmlNodePtr cur;
  char* content;
  int r;

  cur = ablock->cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_get_name_notempty( docinfo->doc, cur, name, &content ) ) {
      r = atoi( content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

      return r;
    }
    cur = cur->next;
  }
  return 0;
}

long mpn_app_block_get_long( mpn_docinfo_ptr docinfo,mpn_app_block_ptr ablock, 
			     const char* name )
{
  xmlNodePtr cur;
  char* content;
  long r;

  cur = ablock->cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_get_name_notempty( docinfo->doc, cur, name, &content ) ) {
      r = atol( content );
      
#if (defined (__UNIX__))
	  free( content );
#endif
      return r;
    }
    cur = cur->next;
  }
  return 0;
}

void mpn_app_block_get_string( mpn_docinfo_ptr docinfo, 
			       mpn_app_block_ptr ablock, 
			       const char* name, char* val )
{
  xmlNodePtr cur;
  char* content;

  cur = ablock->cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_get_name_notempty( docinfo->doc, cur, name, &content ) ) {
      strcpy( val, content );

#if (defined (__UNIX__))
	  free( content );
#endif

      return;
    }
    cur = cur->next;
  }
}

int mpn_app_block_get_filter( mpn_docinfo_ptr docinfo, 
			      mpn_app_block_ptr ablock, 
			      const char* name, mpn_filter_ptr* filter )
{
  xmlNodePtr cur;
  mpn_filter_ptr f;

  cur = ablock->cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_has_name_notempty( docinfo->doc, cur, name ) ) {
      f = parse_filter( docinfo->doc, cur );
      if( f ) {
	*filter = f;
	return MPN_OK;
      }
    }
    cur = cur->next;
  }
  return MPN_PERM_ERR;
}

void mpn_app_block_set_int( mpn_docinfo_ptr docinfo, mpn_app_block_ptr ablock,
			    const char* name, int i )
{
  char buf[MPN_STRING_LEN];
  sprintf( buf, "%d", i );

  delete_node( docinfo->doc, ablock->cur, name );
  xmlNewChild( ablock->cur, NULL, name, buf );
}

void mpn_app_block_set_long( mpn_docinfo_ptr docinfo, mpn_app_block_ptr ablock,
			     const char* name, long l )
{
  char buf[MPN_STRING_LEN];
  sprintf( buf, "%ld", l );

  delete_node( docinfo->doc, ablock->cur, name );
  xmlNewChild( ablock->cur, NULL, name, buf );
}

void mpn_app_block_set_string(mpn_docinfo_ptr docinfo,mpn_app_block_ptr ablock,
			      const char* name, const char* s )
{
  delete_node( docinfo->doc, ablock->cur, name );
  xmlNewChild( ablock->cur, NULL, name, s );
}

int mpn_app_block_get_subblock_count( mpn_docinfo_ptr docinfo, 
				      mpn_app_block_ptr ablock,
				      const char* name )
{
  xmlNodePtr cur;
  int count=0;

  cur = ablock->cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_has_name( cur, name ) ) {
      count++;
    }
    cur = cur->next;
  }
  return count;
}

int mpn_app_block_get_subblock( mpn_docinfo_ptr docinfo, 
				mpn_app_block_ptr ablock,
				const char* name, int i,
				mpn_app_block_ptr* subblock)
{
  xmlNodePtr cur;
  int count=0;
  mpn_app_block_ptr sb;

  cur = ablock->cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_has_name( cur, name ) ) {
      if( i!=count ) count++;
      else {
	sb = calloc( sizeof( mpn_app_block ), 1 );
	strcpy( sb->name, name );
	sb->cur = cur;
	*subblock = sb;
	return MPN_OK;
      }
    }
    cur = cur->next;
  }
  return MPN_PERM_ERR;  
}

void mpn_app_block_free( mpn_app_block_ptr subblock )
{
  free( subblock );
}

/* PREFS PREFS PREFS PREFS PREFS */

/* Dealing with prefs files */

/* Internal functions */

void free_identity( mpn_identity_ptr id ) 
{
  if( id->person )
  {
	  free( id->person );
  }
  if( id->local )
  {
	  free( id->local );
  }
  if( id->remote )
  {
	  free( id->remote );
  }
  free( id );
}

void free_friend( mpn_friend_ptr frienditem ) 
{
  if( frienditem->person ) free( frienditem->person );
  if( frienditem->filter ) mpn_filter_free( frienditem->filter );
  free( frienditem );
}

mpn_friend_ptr parse_friend( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_friend_ptr frienditem;
  mpn_person_ptr personitem;
  mpn_filter_ptr filteritem;
  char* content;

  frienditem = (mpn_friend_ptr)calloc( sizeof( mpn_friend ), 1 );
  frienditem->poll = 10;
  frienditem->cur = cur;

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {

    } else if( node_has_name_notempty( doc, cur, "person" ) ) {
      personitem = parse_person( doc, cur );
      if( personitem ) {
	if( frienditem->person ) free( frienditem->person );
	frienditem->person = personitem;
      }

    } else if( node_get_name_notempty( doc, cur, "poll", (char**)&content ) ) {
      frienditem->poll = atoi( content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_has_name( cur, "bot" ) ) {
      frienditem->bot = 1;

    } else if( node_get_name_notempty( doc, cur, "cookie", (char**)&content ) ) {
      strcpy( frienditem->cookie, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_has_name_notempty( doc, cur, "filter" ) ) {
      filteritem = parse_filter( doc, cur );
      if( filteritem ) {
	if( frienditem->filter ) free( frienditem->filter );
	frienditem->filter = filteritem;
      }

    }
    cur = cur->next;
  }  

  if( frienditem->person == 0 ) {
    free_friend( frienditem );
    return NULL;
  }

  return frienditem;
}

mpn_id_name_ptr parse_idname( xmlDocPtr doc, xmlNodePtr cur, int* def )
{
  mpn_id_name_ptr idname;
  char* content;

  idname = (mpn_id_name_ptr)calloc( sizeof( mpn_id_name ), 1 );
  *def = 0;
  idname->cur = cur;

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_get_name_notempty( doc, cur, "id", &content ) ) {
      idname->id = atoi( content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "name", &content ) ) {
      strcpy( idname->name, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_has_name( cur, "default" ) ) {
      *def = 1;

    }
    cur = cur->next;
  }  

  return idname;
}

mpn_id_name_list_ptr parse_idnamelist( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_id_name_list_ptr idnamelist;
  char* content;
  mpn_id_name_ptr idname;
  int def;

  content = xmlGetProp( cur, "name" );
  if( !content ) return NULL;
  
  idnamelist = (mpn_id_name_list_ptr)calloc( sizeof( mpn_id_name_list ), 1 );
  strcpy( idnamelist->name, content );

#if (defined (__UNIX__))
  free( content );
#endif

  idnamelist->cur = cur;

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_has_name_notempty( doc, cur, "entry" ) ) {
      idname = parse_idname( doc, cur, &def );
      if( idname ) {
	if( def ) idnamelist->def = idnamelist->len;
	idnamelist->list = (mpn_id_name_ptr*)realloc( 
                     idnamelist->list, 
		     sizeof( mpn_id_name_ptr ) * (idnamelist->len+1) );
	idnamelist->list[idnamelist->len] = idname;
	idnamelist->len++;
      }

    }
    cur = cur->next;
  }  

  return idnamelist;
}

mpn_filter_ptr parse_filter( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_filter_ptr filter;
  char* content;

  filter = (mpn_filter_ptr)calloc( sizeof( mpn_filter ), 1 );
  
  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_get_name_notempty( doc, cur, "xslt_proc", &content ) ) {
      strcpy( filter->xslprocess, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    }
    cur = cur->next;
  }  

  return filter;
}

mpn_filter_ptr parse_deffriendfilter( xmlDocPtr doc, xmlNodePtr cur ) 
{
  cur = cur->xmlChildrenNode;

  if( node_has_name_notempty( doc, cur, "filter" ) ) {
    return parse_filter( doc, cur );
  }
  return NULL;
}

mpn_local_account_ptr parse_local_account( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_local_account_ptr local;
  char* content;

  local = calloc( sizeof( mpn_local_account ), 1 );
  
  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_get_name_notempty( doc, cur, "news_file", &content ) ) {
      strcpy( local->news_file, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "space_dir", &content ) ) {
      strcpy( local->space_dir, content );
      
#if (defined (__UNIX__))	  
	  free( content );
#endif

    }
    cur = cur->next;
  }  

  if( local->news_file[0] && local->space_dir[0] ) {
    return local;
  } else {
    free( local );
    return NULL;
  }
}

mpn_remote_account_ptr parse_remote_account( xmlDocPtr doc, xmlNodePtr cur )
{
  mpn_remote_account_ptr remote;
  char* content;

  remote = calloc( sizeof( mpn_remote_account ), 1 );
  strcpy( remote->mailserver, "localhost" );

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_get_name_notempty( doc, cur, "userid", &content ) ) {
      strcpy( remote->userid, content );
      
#if (defined (__UNIX__))	  
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "email", &content ) ) {
      strcpy( remote->email, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "mailserver", &content ) ) {
      strcpy( remote->mailserver, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_get_name_notempty( doc, cur, "server_email", &content ) ) {
      strcpy( remote->server_email, content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_has_name( cur, "mailreply" ) ) {
      remote->mailreply = 1;
    }
    cur = cur->next;
  }  

  if( remote->email[0] && remote->server_email[0] && remote->userid[0] ) {
    return remote;
  } else {
    free( remote );
    return NULL;
  }
}

mpn_identity_ptr parse_identity( xmlDocPtr doc, xmlNodePtr cur, int* def )
{
  mpn_identity_ptr id;
  mpn_person_ptr person;
  mpn_local_account_ptr local;
  mpn_remote_account_ptr remote;
  char* content;

  *def = 0;
  id = calloc( sizeof( mpn_identity ), 1 );
  id->cur = cur; 

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {
    } else if( node_has_name_notempty( doc, cur, "person" ) ) {
      person = parse_person( doc, cur );
      if( person ) {
	if( id->person ) free( id->person );
	id->person = person;
      }

    } else if( node_has_name_notempty( doc, cur, "local" ) ) {
      local = parse_local_account( doc, cur );
      if( local ) {
	if( id->local ) free( id->local );
	id->local = local;
      }

    } else if( node_has_name_notempty( doc, cur, "remote" ) ) {
      remote = parse_remote_account( doc, cur );
      if( remote ) {
	if( id->remote ) free( id->remote );
	id->remote = remote;
      }

    } else if( node_get_name_notempty( doc, cur, "public_space", &content ) ) {
      strcpy( id->public_space, content );

#if (defined (__UNIX__))
	  free( content );
#endif

    } else if( node_has_name( cur, "default" ) ) {
      *def = 1;

    } else if( node_get_name_notempty( doc, cur, "reserve", &content ) ) {
      id->reserve = atoi( content );
      
#if (defined (__UNIX__))
	  free( content );
#endif

    }
    cur = cur->next;
  }  

  if( id->person && (id->local || id->remote ) && id->public_space[0] ) {
    return id;
  } else {
    free_identity( id );
    return NULL;
  }
}

int open_prefs_file( const char* filename, const char* root, 
		     mpn_prefs_ptr* prefsdoc_ret )
{
  xmlDocPtr doc;
  xmlNodePtr cur;
  char* content;
  mpn_prefs_ptr prefsdoc;
  mpn_friend_ptr frienditem;
  mpn_id_name_list_ptr idnamelist;
  mpn_filter_ptr filter;
  mpn_identity_ptr iditem;
  mpn_app_ptr appitem;
  int def;

  /* Open File */
  doc = xmlParseFile( filename );
  if( doc==NULL ) {
    fprintf( stderr, "Error: Prefsfile %s doesnt exist.\n", filename );
    return MPN_PERM_ERR;
  }
  cur = xmlDocGetRootElement( doc );
  if (cur == NULL) {
    fprintf(stderr,"Error: empty document %s.\n", filename );
    xmlFreeDoc(doc);
    return MPN_PERM_ERR;
  }
  if (xmlStrcmp(cur->name, (const xmlChar *)root )) {
    fprintf(stderr,"Error: document %s of the wrong type, root node != %s", 
	    filename, root );
    xmlFreeDoc(doc);
    return MPN_PERM_ERR;
  }

  /* Build mpn_prefs */
  prefsdoc = calloc( sizeof( mpn_prefs ), 1 );
  prefsdoc->mode = MPN_OPEN_READONLY;
  prefsdoc->doc.doc = doc;
  prefsdoc->cur = cur;

  prefsdoc->identities = calloc( sizeof( mpn_identity_list ), 1 );
  prefsdoc->friends = calloc( sizeof( mpn_friend_list ), 1 );
  prefsdoc->id_name_lists = calloc( sizeof( mpn_all_id_name_lists ), 1 );

  cur = cur->xmlChildrenNode;
  while( cur ) {

    if( xmlIsBlankNode( cur ) ) {

    } else if( node_has_name_notempty( doc, cur, "identity" ) ) {
      iditem = parse_identity( doc, cur, &def );
      if( iditem ) {
	prefsdoc->identities->list = (mpn_identity_ptr*)realloc( 
                    prefsdoc->identities->list, 
		    sizeof( mpn_identity_ptr ) * (prefsdoc->identities->len+1) );
	prefsdoc->identities->list[prefsdoc->identities->len] = iditem;
	if( def ) {
	  prefsdoc->identities->def = prefsdoc->identities->len;
	}
	prefsdoc->identities->len++;
      }

    } else if( node_has_name_notempty( doc, cur, "friend" ) ) {
      frienditem = parse_friend( doc, cur );
      if( frienditem ) {
	prefsdoc->friends->list = (mpn_friend_ptr*)realloc( 
                     prefsdoc->friends->list, 
		     sizeof( mpn_friend_ptr ) * (prefsdoc->friends->len+1) );
	prefsdoc->friends->list[prefsdoc->friends->len] = frienditem;
	prefsdoc->friends->len++;
      }

    } else if( node_has_name( cur, "id_name_list" ) ) {
      idnamelist = parse_idnamelist( doc, cur );
      if( idnamelist ) {
	prefsdoc->id_name_lists->list = (mpn_id_name_list_ptr*)realloc( 
             prefsdoc->id_name_lists->list, 
	     sizeof( mpn_id_name_list_ptr ) * (prefsdoc->id_name_lists->len+1) );
	prefsdoc->id_name_lists->list[prefsdoc->id_name_lists->len]=idnamelist;
	prefsdoc->id_name_lists->len++;
      }

    } else if( node_has_name_notempty( doc, cur, "filter" ) ) {
      filter = parse_filter( doc, cur );
      if( filter ) {
	if( prefsdoc->filter ) free( prefsdoc->filter );
	prefsdoc->filter = filter;
      }

    } else if( node_has_name_notempty( doc, cur, "default_friendfilter" ) ) {
      filter = parse_deffriendfilter( doc, cur );
      if( filter ) {
	if( prefsdoc->def_friendfilter ) free( prefsdoc->def_friendfilter );
	prefsdoc->def_friendfilter = filter;
      }

    } else if( node_has_name_notempty( doc, cur, "default_botfilter" ) ) {
      filter = parse_deffriendfilter( doc, cur );
      if( filter ) {
	if( prefsdoc->def_botfilter ) free( prefsdoc->def_botfilter );
	prefsdoc->def_botfilter = filter;
      }

    } else if( node_has_name_notempty( doc, cur, "app" ) ) {
      appitem = parse_app( doc, cur );
      if( appitem ) {
	if( prefsdoc->app ) free_app( prefsdoc->app );
	prefsdoc->app = appitem;
      }

    } else if( node_get_name_notempty( doc, cur, "default_location", &content ) ) {
      strcpy( prefsdoc->deflocation, content );
	  
#if (defined (__UNIX__))
	  free( content );
#endif

    }

    cur = cur->next;
  }

  *prefsdoc_ret = prefsdoc;
  return MPN_OK;
}

int mpn_prefs_open_file_once( const char* filename, int mode, 
			      mpn_prefs_ptr* prefsdoc_ret )
{
  int r;
  mpn_prefs_ptr prefsdoc;

  /* Do Lock Stuff */
  if( mode==MPN_OPEN_READONLY ) {
    r = check_lock( filename );
    if( r!=MPN_OK ) return r;

  } else if( mode==MPN_OPEN_READWRITE ) {
    r = lock_file( filename );
    if( r!=MPN_OK ) return r;

  } else {
    return MPN_PERM_ERR;
  }

  /* Open and Read */
  r = open_prefs_file( filename, "matrix-private-prefs", &prefsdoc);

  /* Everything good? */
  if( r!=MPN_OK ) {
    if( mode==MPN_OPEN_READWRITE ) unlock_file( filename );
    return r;

  } else {
    prefsdoc->mode = mode;
    strcpy( prefsdoc->filename, filename ) ;
    *prefsdoc_ret = prefsdoc;
    return MPN_OK;
  }
}

/* External functions */

int mpn_prefs_open_private( int mode, mpn_prefs_ptr* prefsdoc_ret )
{
  int i, r;
  mpn_prefs_ptr prefsdoc;
//#ifdef LINUX
  //struct timespec wt;
//#endif
   
  for( i=0, r=MPN_TEMP_ERR; (i<10)&&(r!=MPN_OK); i++ ) {
    r = mpn_prefs_open_file_once( mpn_get_prefsfile(), mode, &prefsdoc );
    if( r!=MPN_OK ) {
//#ifdef LINUX
      //wt.tv_sec=0;
      //wt.tv_nsec=100000000;
      //nanosleep( &wt, NULL );
//#endif
//#ifdef WIN32
	  //Sleep(100000);
//#endif
		mfl_sleep(500);
      printf( "open prefs retry %d\n", 1 );
    }
  }

  if( r==MPN_OK ) {
    *prefsdoc_ret = prefsdoc;
    return MPN_OK;
  } else {
    return MPN_PERM_ERR;
  }
}

void mpn_prefs_close( mpn_prefs_ptr prefsdoc )
{
  int i, j;

  if( prefsdoc->mode==MPN_OPEN_READWRITE ) {
    xmlSaveFormatFile( prefsdoc->filename, prefsdoc->doc.doc, 1 );
    unlock_file( prefsdoc->filename );
  }

  for( i=0; i<prefsdoc->identities->len; i++ ) {
    free_identity( prefsdoc->identities->list[i] );
  }
  free( prefsdoc->identities->list );

  xmlFreeDoc( prefsdoc->doc.doc );
  for( i=0; i<prefsdoc->friends->len; i++ ) {
    free_friend( prefsdoc->friends->list[i] );
  }
  free( prefsdoc->friends->list );

  for( i=0; i<prefsdoc->id_name_lists->len; i++ ) {
    for( j=0; j<prefsdoc->id_name_lists->list[i]->len; j++ ) {
      free( prefsdoc->id_name_lists->list[i]->list[j] );
    }
    free( prefsdoc->id_name_lists->list[i]->list );
    free( prefsdoc->id_name_lists->list[i] );
  }
  free( prefsdoc->id_name_lists->list );
  free( prefsdoc->id_name_lists );

  if( prefsdoc->filter )
    free( prefsdoc->filter );

  if( prefsdoc->def_friendfilter )
    free( prefsdoc->def_friendfilter );
  if( prefsdoc->def_botfilter )
    free( prefsdoc->def_botfilter );

  if( prefsdoc->app ) 
    free_app( prefsdoc->app );

  free( prefsdoc->friends );
  free( prefsdoc );
}

/* Dealing with prefs */

mpn_docinfo_ptr mpn_prefs_get_docinfo( mpn_prefs_ptr prefsdoc )
{
  return &prefsdoc->doc;
}

int mpn_prefs_get_identity_count( mpn_prefs_ptr prefsdoc )
{
  return prefsdoc->identities->len;
}

int mpn_prefs_get_identity_default( mpn_prefs_ptr prefsdoc )
{
  return prefsdoc->identities->def;
}

mpn_identity_ptr mpn_prefs_get_identity( mpn_prefs_ptr prefsdoc, int i )
{
  if( i<prefsdoc->identities->len ) {
    return prefsdoc->identities->list[i];
  } else {
    return NULL;
  }
}

int mpn_prefs_del_identity( mpn_prefs_ptr prefsdoc, int i )
{
  mpn_identity_ptr exid;
  int j;

  exid = mpn_prefs_get_identity( prefsdoc, i );
  xmlUnlinkNode( exid->cur );
  xmlFreeNode( exid->cur );      
  free_identity( exid );

  for( j=i; j<prefsdoc->identities->len-1; j++ ) {
    prefsdoc->identities->list[j] = prefsdoc->identities->list[j+1];
  }

  prefsdoc->identities->list = realloc( prefsdoc->identities->list, 
		 sizeof( mpn_identity_ptr ) * (prefsdoc->identities->len-1) );
  prefsdoc->identities->len--;

  return MPN_OK;
}

int mpn_prefs_add_remote_identity( mpn_prefs_ptr prefsdoc , const char* name, 
				   const char* home, const char* userid,
				   const char* email, const char* server_email,
				   const char* public_space,
				   const char* mailserver )
{
  mpn_identity_ptr id;
  mpn_person_ptr person;
  mpn_remote_account_ptr remote;
  xmlNodePtr personcur;
  xmlNodePtr remotecur;

  // Make internal Structure
  person = calloc( sizeof( mpn_person ), 1 );
  strcpy( person->name, name );
  strcpy( person->home, home );

  remote = calloc( sizeof( mpn_remote_account ), 1 );
  strcpy( remote->userid, userid );
  strcpy( remote->email, email );
  strcpy( remote->mailserver, mailserver );
  strcpy( remote->server_email, server_email );

  id = calloc( sizeof( mpn_identity ), 1 );
  id->person = person;
  id->remote = remote;
  strcpy( id->public_space, public_space );

  // Add to internal Structure
  prefsdoc->identities->list = realloc( prefsdoc->identities->list, 
		 sizeof( mpn_identity_ptr ) * (prefsdoc->identities->len+1) );
  prefsdoc->identities->list[prefsdoc->identities->len] = id;
  prefsdoc->identities->len++;
  
  // Make XML Structure
  id->cur = xmlNewChild( prefsdoc->cur, NULL, "identity", NULL );

  personcur = xmlNewChild( id->cur, NULL, "person", NULL );
  xmlNewChild( personcur, NULL, "name", name );
  xmlNewChild( personcur, NULL, "home", home );

  remotecur = xmlNewChild( id->cur, NULL, "remote", NULL );
  xmlNewChild( remotecur, NULL, "userid", userid );
  xmlNewChild( remotecur, NULL, "email", email );
  xmlNewChild( remotecur, NULL, "mailserver", mailserver );
  xmlNewChild( remotecur, NULL, "server_email", server_email );

  xmlNewChild( id->cur, NULL, "public_space", public_space );

  return MPN_OK;
}

int mpn_prefs_get_friend_count( mpn_prefs_ptr prefsdoc )
{
  return prefsdoc->friends->len;
}

mpn_friend_ptr mpn_prefs_get_friend( mpn_prefs_ptr prefsdoc, int i )
{
  if( i<prefsdoc->friends->len ) {
    return prefsdoc->friends->list[i];
  } else {
    return NULL;
  }
}

int mpn_prefs_add_friend( mpn_prefs_ptr prefsdoc , const char* name, 
			  const char* home, int poll, int bot, 
			  const char* cookie )
{
  mpn_friend_ptr friend;
  mpn_person_ptr person;
  xmlNodePtr personcur;
  char buf[MPN_STRING_LEN];

  // Make internal Structure
  friend = calloc( sizeof( mpn_friend ), 1 );
  person = calloc( sizeof( mpn_person ), 1 );
  strcpy( person->name, name );
  strcpy( person->home, home );
  friend->person = person;
  friend->poll = poll;
  friend->bot = bot;
  if( cookie ) 
    strcpy( friend->cookie, cookie );

  // Add to internal Structure
  prefsdoc->friends->list = (mpn_friend_ptr*)realloc( 
		     prefsdoc->friends->list, 
		     sizeof( mpn_friend_ptr ) * (prefsdoc->friends->len+1) );
  prefsdoc->friends->list[prefsdoc->friends->len] = friend;
  prefsdoc->friends->len++;
  
  // Make XML Structure
  friend->cur = xmlNewChild( prefsdoc->cur, NULL, "friend", NULL );
  personcur = xmlNewChild( friend->cur, NULL, "person", NULL );
  xmlNewChild( personcur, NULL, "name", name );
  xmlNewChild( personcur, NULL, "home", home );
  if( bot ) {
    xmlNewChild( friend->cur, NULL, "bot", NULL );
  }
  if( cookie ) {
    xmlNewChild( friend->cur, NULL, "cookie", cookie );
  }

  sprintf( buf, "%d", poll );
  xmlNewChild( friend->cur, NULL, "poll", buf );

  return MPN_OK;
}

int mpn_prefs_del_friend( mpn_prefs_ptr prefsdoc, int i )
{
  mpn_friend_ptr exfriend;
  int j;

  exfriend = mpn_prefs_get_friend( prefsdoc, i );
  xmlUnlinkNode( exfriend->cur );
  xmlFreeNode( exfriend->cur );      
  free_friend( exfriend );

  for( j=i; j<prefsdoc->friends->len-1; j++ ) {
    prefsdoc->friends->list[j] = prefsdoc->friends->list[j+1];
  }

  prefsdoc->friends->list = realloc( prefsdoc->friends->list, 
		     sizeof( mpn_friend_ptr ) * (prefsdoc->friends->len-1) );
  prefsdoc->friends->len--;

  return MPN_OK;
}

mpn_id_name_list_ptr mpn_prefs_get_idnamelist( mpn_prefs_ptr prefsdoc, 
					       const char* name )
{
  int i;

  for( i=0; i<prefsdoc->id_name_lists->len; i++ ) {
    if( strcmp( name, prefsdoc->id_name_lists->list[i]->name )==0 ) {
      return prefsdoc->id_name_lists->list[i];
    }
  }
  return NULL;
}
mpn_id_name_list_ptr mpn_prefs_add_idnamelist( mpn_prefs_ptr prefsdoc,
					       const char* name )
{
  mpn_id_name_list_ptr idnamelist;

  // Make internal Structure
  idnamelist = calloc( sizeof( mpn_id_name_list ), 1 );
  strcpy( idnamelist->name, name );

  // Add to internal Structure
  prefsdoc->id_name_lists->list = realloc( prefsdoc->id_name_lists->list, 
	  sizeof( mpn_id_name_list_ptr ) * (prefsdoc->id_name_lists->len+1) );
  prefsdoc->id_name_lists->list[prefsdoc->id_name_lists->len] = idnamelist;
  prefsdoc->id_name_lists->len++;
  
  // Make XML Structure
  idnamelist->cur = xmlNewChild( prefsdoc->cur, NULL, "id_name_list", NULL );
  xmlSetProp( idnamelist->cur, "name", name );

  return idnamelist;
}

mpn_filter_ptr mpn_prefs_get_globalfilter( mpn_prefs_ptr prefsdoc )
{
  return prefsdoc->filter;
}

mpn_filter_ptr mpn_prefs_get_deffriendfilter( mpn_prefs_ptr prefsdoc )
{
  return prefsdoc->def_friendfilter;
}

mpn_filter_ptr mpn_prefs_get_defbotfilter( mpn_prefs_ptr prefsdoc )
{
  return prefsdoc->def_botfilter;
}

mpn_app_ptr mpn_prefs_get_app( mpn_prefs_ptr prefsdoc )
{
  return prefsdoc->app;
}

const char* mpn_prefs_get_defaultlocation( mpn_prefs_ptr prefsdoc )
{
  return prefsdoc->deflocation;
}

/* Dealing with identities */

mpn_person_ptr mpn_identity_get_person( mpn_identity_ptr id )
{
  return id->person;
}

mpn_local_account_ptr mpn_identity_get_local_account( mpn_identity_ptr id )
{
  return id->local;
}

mpn_remote_account_ptr mpn_identity_get_remote_account( mpn_identity_ptr id )
{
  return id->remote;
}

const char* mpn_identity_get_public_space( mpn_identity_ptr id )
{
  return id->public_space;
}

int mpn_identity_get_reserve( mpn_identity_ptr id )
{
  return id->reserve;
}

/* Dealing with local accounts */

const char* mpn_local_account_get_news( mpn_local_account_ptr la )
{
  return la->news_file;
}

const char* mpn_local_account_get_space( mpn_local_account_ptr la )
{
  return la->space_dir;
}

/* Dealing with remote accounts */

const char* mpn_remote_account_get_userid( mpn_remote_account_ptr ra )
{
  return ra->userid;
}

const char* mpn_remote_account_get_email( mpn_remote_account_ptr ra )
{
  return ra->email;
}

const char* mpn_remote_account_get_mailserver( mpn_remote_account_ptr ra )
{
  return ra->mailserver;
}

const char* mpn_remote_account_get_serveremail( mpn_remote_account_ptr ra )
{
  return ra->server_email;
}

int mpn_remote_account_get_mailreply( mpn_remote_account_ptr ra )
{
  return ra->mailreply;
}

/* Dealing with friends */

mpn_person_ptr mpn_friend_get_person( mpn_friend_ptr friend )
{
  return friend->person;
}

int mpn_friend_get_poll( mpn_friend_ptr friend )
{
  return friend->poll;
}

const char* mpn_friend_get_cookie( mpn_friend_ptr friend )
{
  return friend->cookie;
}

mpn_filter_ptr mpn_friend_get_filter( mpn_friend_ptr friend )
{
  return friend->filter;
}

int mpn_friend_is_bot( mpn_friend_ptr friend )
{
  return friend->bot;
}

void mpn_friend_set_cookie( mpn_prefs_ptr prefsdoc, mpn_friend_ptr friend, 
			    const char* cookie )
{
  strcpy( friend->cookie, cookie );

  delete_node( prefsdoc->doc.doc, friend->cur, "cookie" );
  xmlNewChild( friend->cur, NULL, "cookie", friend->cookie );
}

/* Dealing with ID-Name Lists */

int mpn_idnamelist_get_len( mpn_id_name_list_ptr idnamelist )
{
  return idnamelist->len;
}

int mpn_idnamelist_get_default( mpn_id_name_list_ptr idnamelist )
{
  return idnamelist->def;
}

void mpn_idnamelist_set_default( mpn_prefs_ptr prefsdoc, 
				 mpn_id_name_list_ptr idnamelist, int i )
{
  mpn_id_name_ptr idname;

  idname = idnamelist->list[ idnamelist->def ];
  delete_node( prefsdoc->doc.doc, idname->cur, "default" );

  idnamelist->def = i;
  idname = idnamelist->list[ idnamelist->def ];
  xmlNewChild( idname->cur, NULL, "default", NULL );
}

mpn_id_name_ptr mpn_idnamelist_get_idname( mpn_id_name_list_ptr idnamelist, 
					   int i )
{
  return idnamelist->list[i];
}

void mpn_idnamelist_add_entry( mpn_id_name_list_ptr idnamelist, int id,
			       const char* name )
{
  mpn_id_name_ptr idname;
  char buf[MPN_STRING_LEN];

  // Make internal Structure
  idname = calloc( sizeof( mpn_id_name ), 1 );
  idname->id = id;
  strcpy( idname->name, name );

  // Add to internal Structure
  idnamelist->list = realloc( idnamelist->list, 
	  sizeof( mpn_id_name_ptr ) * (idnamelist->len+1) );
  idnamelist->list[idnamelist->len] = idname;
  idnamelist->len++;
  
  // Make XML Structure
  idname->cur = xmlNewChild( idnamelist->cur, NULL, "entry", NULL );
  sprintf( buf, "%d", id );
  xmlNewChild( idname->cur, NULL, "id", buf );
  xmlNewChild( idname->cur, NULL, "name", name );
}

void mpn_idnamelist_del_entry( mpn_id_name_list_ptr idnamelist, int i )
{
  mpn_id_name_ptr exidname;
  int j;

  exidname = mpn_idnamelist_get_idname( idnamelist, i );
  xmlUnlinkNode( exidname->cur );
  xmlFreeNode( exidname->cur );      
  free( exidname );

  for( j=i; j<idnamelist->len-1; j++ ) {
    idnamelist->list[j] = idnamelist->list[j+1];
  }

  idnamelist->list = realloc( idnamelist->list, 
		 sizeof( mpn_id_name_ptr ) * (idnamelist->len-1) );
  idnamelist->len--;
}

/* Dealing with ID-Names */

int mpn_idname_get_id( mpn_id_name_ptr idname )
{
  return idname->id;
}

const char* mpn_idname_get_name( mpn_id_name_ptr idname )
{
  return idname->name;
}

/* Dealing with filters */

const char* mpn_filter_get_xslprocess( mpn_filter_ptr f )
{
  return f->xslprocess;
}

void mpn_filter_free( mpn_filter_ptr filter )
{
  free( filter );
}




//////////////////////////////////////////////////////
// MFL

#include <stdio.h>

#if (defined (__WINDOWS__))
#include <windows.h>
#include <direct.h>
#endif

#if (defined (__UNIX__))
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

void mfl_sleep(int msec)
{
#if (defined (__UNIX__))
  struct timespec wt;
  wt.tv_sec=msec/1000;
  wt.tv_nsec=1000 * 1000 * (msec%1000);
  nanosleep( &wt, NULL );
#endif

#if (defined (__WINDOWS__))
  Sleep(msec);    
#endif
}

int mfl_mkdir(char* dirname)
{
#if (defined (__UNIX__))
	mkdir(dirname, 0);
	chmod( dirname, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP );
	return 0;
#endif
#if (defined (__WINDOWS__))
	return mkdir(dirname);
#endif
}





