/*
** Copyright (C) 2002 Tom Nicolai <nicolai@tzi.de>, Nils Behrens <psi@tzi.de>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "mixin.h"
#include <vorbis/vorbisfile.h>
#include <time.h>
#include <mpn/matrix.h>

#if (defined (__UNIX__))
#include <unistd.h>
#endif

#if (defined (__WINDOWS__))
#include <memory.h>
#include <string.h>
#endif


struct _audio_file_t {
  char filename[MPN_STRING_LEN];
  int open;
  AFormat afmt;
  int srate;
  int nch;

  FILE* file;
  OggVorbis_File ogg_file;

  void(*open_function)(void);
  int(*read_function)(char*,int);
  void(*close_function)(void);

} audio_file;
  
struct _message_t {
  char id[MPN_STRING_LEN];
  char type[MPN_STRING_LEN];
  char filename[MPN_STRING_LEN];
  int duration;
  double weight;
};
typedef struct _message_t message_t;
typedef message_t* message_t_ptr;

struct _message_list {
  int pos;
  int len;
  message_t_ptr* list;
  int d_time;
};
typedef struct _message_list message_list;
typedef message_list* message_list_ptr;

message_list_ptr play_list;

#define MAX_NEWSBLOCKS 20

struct _newsblock_list {
  int len;
  long lastcheck[MAX_NEWSBLOCKS];
};
typedef struct _newsblock_list newsblock_list;
typedef newsblock_list* newsblock_list_ptr;

newsblock_list newsblocks;

#define WAV_BUF_SIZE (16384*4)
char wav_buf[WAV_BUF_SIZE];
int wav_buf_pos;
int wav_buf_full;
int current_section;

int wav_read( char* buf, int len, AFormat afmt, int srate, int nch );
int dec_read( char* buf, int len, AFormat afmt, int srate, int nch );

void ogg_open();
int ogg_read( char* buf, int len );
void ogg_close();

void prepare_next_audio_file( message_list_ptr ml );
void free_message_list( message_list_ptr ml );
void setplayed_message_list( message_list_ptr ml );
void sort_message_list( message_list_ptr ml );
int make_new_message_list( mpn_app_block_ptr nblock, message_list_ptr* ml_ret);

// is regularly set when xmms is playing
int playing;
// playing news
int newsing;

/* External Functions */

void mixin_init()
{
  audio_file.open = 0;
  wav_buf_full = 0;
  wav_buf_pos = 0;
  playing = 0;
  newsing = 0;
  play_list = NULL;
  newsblocks.len = 0;
}

void mixin_cleanup()
{
  if( play_list ) 
    free_message_list( play_list );
}

int samples_available()
{
  playing = 1;
  return newsing;
}

int get_samples( char *buf, int len, AFormat afmt, int srate, int nch )
{
  char* wav_buf_ptr;
  int n;
  char* dest;
  int destleft;

  // Read until the buffer is full, or the queue is empty

  dest = buf;
  destleft = len;

  while( destleft!=0 ) {

    if( wav_buf_full<=wav_buf_pos ) {
      wav_buf_full = wav_read( wav_buf, WAV_BUF_SIZE, afmt, srate, nch );
      wav_buf_pos = 0;
      if( wav_buf_full==0 ) {
	newsing = 0;
	setplayed_message_list( play_list );
	free_message_list( play_list );
	play_list = NULL;
	return len-destleft;
      }
    }
 
    wav_buf_ptr = wav_buf;
    wav_buf_ptr += wav_buf_pos;

    n = wav_buf_full - wav_buf_pos;
    if( n>destleft ) n = destleft;

    memcpy( dest, wav_buf_ptr, n );

    wav_buf_pos += n;
    destleft -= n;
    dest += n;
  }
  
  return len;
}

/* Internal Functions */

int wav_read( char* buf, int len, AFormat afmt, int srate, int nch )
{
  int read;

  // Read twice, in case the audio file is finished,
  // but there is another one in the queue

  read = dec_read( buf, len, afmt, srate, nch );
  if( read==0 ) 
    read = dec_read( buf, len, afmt, srate, nch );

  return read;
}

int dec_1_to_1( char* newbuf, char* buf, int len )
{
  int read;

  read = audio_file.read_function( newbuf, len );
  memcpy( buf, newbuf, read );

  return read;
}

int dec_1_to_2( char* newbuf, char* buf, int len )
{
  int read, i, j, newlen;
  signed short* dest;
  signed short* src;

  read = audio_file.read_function( newbuf, len/2 );
  newlen = read*2;

  dest = (signed short*)buf;
  src  = (signed short*)newbuf;
  for( i=0, j=0; i+1<newlen/2; i+=2, j++ ) {
    dest[i] = src[j];
    dest[i+1] = src[j];
  }
  return newlen;
}

int dec_2_to_4( char* newbuf, char* buf, int len )
{
  int read, i, j, newlen;
  signed short* dest;
  signed short* src;

  read = audio_file.read_function( newbuf, len/2 );
  newlen = read*2;

  dest = (signed short*)buf;
  src  = (signed short*)newbuf;
  for( i=0, j=0; i+3<newlen/2; i+=4, j+=2 ) {
    dest[i] = src[j];
    dest[i+1] = src[j+1];
    dest[i+2] = src[j];
    dest[i+3] = src[j+1];
  }
  return newlen;
}

int dec_1_to_4( char* newbuf, char* buf, int len )
{
  int read, i, j, newlen;
  signed short* dest;
  signed short* src;

  read = audio_file.read_function( newbuf, len/4 );
  newlen = read*4;
  //printf( "dec_1_to_4: len=%d, read=%d, newlen=%d\n", len, read, newlen );

  dest = (signed short*)buf;
  src  = (signed short*)newbuf;
  for( i=0, j=0; i+3<newlen/2; i+=4, j++ ) {
    dest[i] = src[j];
    dest[i+1] = src[j];
    dest[i+2] = src[j];
    dest[i+3] = src[j];
  }
  return newlen;
}

int dec_2_to_1( char* newbuf, char* buf, int len )
{
  int read, i, j, newlen;
  signed short* dest;
  signed short* src;
  int avg;

  read = audio_file.read_function( newbuf, len*2 );
  newlen = read/2;
  newlen/=4;
  newlen*=4;

  dest = (signed short*)buf;
  src  = (signed short*)newbuf;
  for( i=0, j=0; i<newlen/2; i++, j+=2 ) {
    avg = (((int)src[j]) + ((int)src[j+1])) / 2;
    dest[i] = avg;
  }
  return newlen;
}

int dec_4_to_1( char* newbuf, char* buf, int len )
{
  int read, i, j, newlen;
  signed short* dest;
  signed short* src;
  int avg;

  read = audio_file.read_function( newbuf, len*4 );
  newlen = read/4;
  newlen/=8;
  newlen*=8;

  dest = (signed short*)buf;
  src  = (signed short*)newbuf;
  for( i=0, j=0; i<newlen/2; i++, j+=4 ) {
    avg = (((int)src[j]) + ((int)src[j+1]) + ((int)src[j+2]) 
	   + ((int)src[j+3])) / 4;
    dest[i] = avg;
  }
  return newlen;
}

int dec_4_to_2( char* newbuf, char* buf, int len )
{
  int read, i, j, newlen;
  signed short* dest;
  signed short* src;
  int avg;

  read = audio_file.read_function( newbuf, len*2 );
  newlen = read/2;
  newlen/=8;
  newlen*=8;

  dest = (signed short*)buf;
  src  = (signed short*)newbuf;
  for( i=0, j=0; i+2<newlen/2; i+=2, j+=4 ) {
    avg = (((int)src[j]) + ((int)src[j+1])) / 2;
    dest[i] = avg;
    avg = (((int)src[j+2]) + ((int)src[j+3])) / 2;
    dest[i+1] = avg;
  }
  return newlen;
}

int dec_2_to_2( char* newbuf, char* buf, int len )
{
  int read, i, newlen;
  signed short* dest;
  signed short* src;
  int avg;

  read = audio_file.read_function( newbuf, len );
  newlen = read;

  dest = (signed short*)buf;
  src  = (signed short*)newbuf;
  for( i=0; i+1<newlen/2; i+=2 ) {
    avg = (((int)src[i]) + ((int)src[i+1])) / 2;
    dest[i] = avg;
    dest[i+1] = avg;
  }
  return newlen;
}

int dec_read( char* buf, int len, AFormat afmt, int srate, int nch )
{
  char newbuf[100000];
  int read=0;

  if( !audio_file.open )
  {
	prepare_next_audio_file( play_list );
    
	if( audio_file.open_function )
	{
		audio_file.open_function();

		if( !audio_file.open )
		{
			return 0;
		}
	} 
	else
	{
      return 0;
    }
  }

  // Convert 22050Hz, Mono -> 44100Hz, Stereo    
  if( (srate==44100) && (audio_file.srate==22050) 
      && (nch==2) && (audio_file.nch==1) ) {
    read = dec_1_to_4( newbuf, buf, len );

  // Convert 44100Hz, Mono -> 44100Hz, Stereo    
  } else if( (srate==44100) && (audio_file.srate==44100) 
      && (nch==2) && (audio_file.nch==1) ) {
    read = dec_1_to_2( newbuf, buf, len );

  // Convert 22050Hz, Stereo -> 44100Hz, Stereo    
  } else if( (srate==44100) && (audio_file.srate==22050) 
      && (nch==2) && (audio_file.nch==2) ) {
    read = dec_2_to_4( newbuf, buf, len );

  // Convert 44100Hz, Stereo -> 44100Hz, Stereo    
  } else if( (srate==44100) && (audio_file.srate==44100) 
      && (nch==2) && (audio_file.nch==2) ) {
    read = dec_1_to_1( newbuf, buf, len );



  // Convert 22050Hz, Mono -> 44100Hz, Mono    
  } else if( (srate==44100) && (audio_file.srate==22050) 
      && (nch==1) && (audio_file.nch==1) ) {
    read = dec_1_to_2( newbuf, buf, len );

  // Convert 44100Hz, Mono -> 44100Hz, Mono    
  } else if( (srate==44100) && (audio_file.srate==44100) 
      && (nch==1) && (audio_file.nch==1) ) {
    read = dec_1_to_1( newbuf, buf, len );

  // Convert 22050Hz, Stereo -> 44100Hz, Mono    
  } else if( (srate==44100) && (audio_file.srate==22050) 
      && (nch==1) && (audio_file.nch==2) ) {
    read = dec_2_to_2( newbuf, buf, len );

  // Convert 44100Hz, Stereo -> 44100Hz, Mono    
  } else if( (srate==44100) && (audio_file.srate==44100) 
      && (nch==1) && (audio_file.nch==2) ) {
    read = dec_2_to_1( newbuf, buf, len );



  // Convert 22050Hz, Mono -> 22050Hz, Stereo    
  } else if( (srate==22050) && (audio_file.srate==22050) 
      && (nch==2) && (audio_file.nch==1) ) {
    read = dec_1_to_2( newbuf, buf, len );

  // Convert 22050Hz, Stereo -> 22050Hz, Stereo    
  } else if( (srate==22050) && (audio_file.srate==22050) 
      && (nch==2) && (audio_file.nch==2) ) {
    read = dec_1_to_1( newbuf, buf, len );

  // Convert 44100Hz, Mono -> 22050Hz, Stereo    
  } else if( (srate==22050) && (audio_file.srate==44100) 
      && (nch==2) && (audio_file.nch==2) ) {
    read = dec_2_to_2( newbuf, buf, len );

  // Convert 44100Hz, Stereo -> 22050Hz, Stereo    
  } else if( (srate==22050) && (audio_file.srate==44100) 
      && (nch==2) && (audio_file.nch==2) ) {
    read = dec_4_to_2( newbuf, buf, len );



  // Convert 22050Hz, Mono -> 22050Hz, Mono    
  } else if( (srate==22050) && (audio_file.srate==22050) 
      && (nch==1) && (audio_file.nch==1) ) {
    read = dec_1_to_1( newbuf, buf, len );

  // Convert 22050Hz, Stereo -> 22050Hz, Mono    
  } else if( (srate==22050) && (audio_file.srate==22050) 
      && (nch==1) && (audio_file.nch==2) ) {
    read = dec_2_to_1( newbuf, buf, len );

  // Convert 44100Hz, Mono -> 22050Hz, Mono    
  } else if( (srate==22050) && (audio_file.srate==44100) 
      && (nch==1) && (audio_file.nch==2) ) {
    read = dec_2_to_1( newbuf, buf, len );

  // Convert 44100Hz, Stereo -> 22050Hz, Mono    
  } else if( (srate==22050) && (audio_file.srate==44100) 
      && (nch==1) && (audio_file.nch==2) ) {
    read = dec_4_to_1( newbuf, buf, len );

  }

  if( read == 0 ) {
    audio_file.close_function();
  }

  return read;
}

/* OGG File Decoding */

void ogg_open()
{
  vorbis_info* info;

  int debugov_open;

#if (defined (__UNIX__))
  audio_file.file = fopen( audio_file.filename, "r" );
#endif
#if (defined (__WINDOWS__))
  audio_file.file = fopen( audio_file.filename, "rb" );
#endif
  
  if( audio_file.file ) {
    //debugov_open=ov_open( audio_file.file, &audio_file.ogg_file, NULL, 0);
	 
	//if(debugov_open>=0)
	
	/*if(debugov_open==OV_EREAD) debugov_open=100; 
	if(debugov_open==OV_ENOTVORBIS) debugov_open=200;
	if(debugov_open==OV_EVERSION) debugov_open=300;
	if(debugov_open==OV_EBADHEADER) debugov_open=400;
	if(debugov_open==OV_EFAULT) debugov_open=500;*/
	
	//if(debugov_open>=0)
	  if( ov_open( audio_file.file, &audio_file.ogg_file, NULL, 0)>=0 )
	{
		
      info = ov_info( &audio_file.ogg_file, -1 );
      audio_file.afmt = FMT_S16_BE;
      audio_file.srate = info->rate;
      audio_file.nch = info->channels;
      audio_file.open = 1;

    } else {
      fclose( audio_file.file );
    }
  }
}

int ogg_read( char* buf, int len )
{
  return ov_read( &audio_file.ogg_file, buf, len, 0, 2, 1, &current_section );
}

void ogg_close()
{
  if( audio_file.open ) {
    ov_clear( &audio_file.ogg_file );
    audio_file.open = 0;
  }
}

/* Everything about the Message List */

message_t_ptr new_message( mpn_docinfo_ptr docinfo, mpn_newsitem_ptr item )
{
  mpn_sig_ptr sig;
  mpn_file_ptr file;
  mpn_net_ptr net;
  mpn_app_block_ptr ablock;
  message_t_ptr m;

  sig = mpn_newsitem_get_sig( item );
  file = mpn_sig_get_file( sig );
  ablock = mpn_file_get_typeblock( file );
  net = mpn_newsitem_get_net( item );
  
  m = calloc( sizeof(message_t), 1 );
  strcpy( m->id, mpn_sig_get_id( sig ) );
  strcpy( m->type, mpn_file_get_type( file ) );
  m->weight = mpn_net_get_weight( net );
  if( ablock ) {
    m->duration = mpn_app_block_get_int( docinfo, ablock, "duration" );
  }

  return m;
}

message_t_ptr new_ogg_jingle_message( const char* filename )
{
  message_t_ptr m;

  m = calloc( sizeof(message_t), 1 );
  strcpy( m->type, "audio/ogg" );
  strcpy( m->filename, filename );

  return m;
}

int fetch_message( message_t_ptr m, mpn_news_ptr newsdoc )
{
  int i, r;
  mpn_newsitem_ptr item;
  mpn_sig_ptr sig;
  mpn_net_ptr net;
  const char* lfn;
  char filename[MPN_STRING_LEN];

  for( i=0; i<mpn_news_get_len( newsdoc ); i++ ) {
    item = mpn_news_get_item( newsdoc, i );
    sig = mpn_newsitem_get_sig( item );
    net = mpn_newsitem_get_net( item );

    if( strcmp( mpn_sig_get_id( sig ), m->id ) == 0 ) {

      lfn = mpn_net_get_localfile( net );
      if( strcmp( lfn, "" )==0 ) {
	printf( "fetching message %s\n", mpn_sig_get_id( sig ) );
	r = mpn_newsitem_fetch_file( newsdoc, item, filename );
	if( r!=MPN_OK ) {
	  printf( "  ERROR\n" );
	  return MPN_PERM_ERR;
	}
	printf( "  OK %s\n", filename );

      } else {
	strcpy( filename, lfn );
	printf( "message already here: %s\n", filename );
      }
      sprintf( m->filename, "%s/%s", mpn_get_contentdir(), filename );
      
      return MPN_OK;
    }
  }

  return MPN_PERM_ERR;
}

void prepare_next_audio_file( message_list_ptr ml )
{
  message_t_ptr m;

  if( ml->pos>=ml->len ) {
    audio_file.open = 0;
    audio_file.open_function = NULL;
    audio_file.read_function = NULL;
    audio_file.close_function = NULL;

  } else {
    m = ml->list[ml->pos++];

    strcpy( audio_file.filename, m->filename );

    if( strcmp( "audio/ogg", m->type )==0 ) {
      audio_file.open_function = ogg_open;
      audio_file.read_function = ogg_read;
      audio_file.close_function = ogg_close;

    } else {
      audio_file.open_function = NULL;
      audio_file.read_function = NULL;
      audio_file.close_function = NULL;
    }

  }
}

void free_message_list( message_list_ptr ml )
{
  int i;

  if( ml->len>0 ) {
    for( i=0; i<ml->len; i++ ) {
      free( ml->list[i] );
    }
    free( ml->list );
  }
  free( ml );
}

void add_tail_message_list( message_list_ptr ml, message_t_ptr m )
{
  ml->list = realloc( ml->list, sizeof(message_t_ptr)*(ml->len+1) );
  ml->list[ml->len] = m;
  ml->len++;
}

void add_pos_message_list( message_list_ptr ml, message_t_ptr m, int p )
{
  int i;

  ml->list = realloc( ml->list, sizeof(message_t_ptr)*(ml->len+1) );
  for( i=ml->len; i>p; i-- ) {
    ml->list[i] = ml->list[i-1];
  }
  ml->len++;
  ml->list[p] = m;
}

void del_pos_message_list( message_list_ptr ml, int p )
{
  int i;

  free( ml->list[p] );
  for( i=p; i<ml->len-1; i++ ) {
    ml->list[i] = ml->list[i+1];
  }
  ml->len--;
}

void setplayed_message_list( message_list_ptr ml )
{
  mpn_news_ptr newsdoc;
  mpn_newsitem_ptr item;
  mpn_app_ptr app;
  mpn_app_block_ptr appblock;
  mpn_sig_ptr sig;
  mpn_docinfo_ptr docinfo;
  int i, j, r;
  time_t timestamp;

  if( ml->len>0 ) {

    r = mpn_news_open_private( MPN_OPEN_READWRITE, &newsdoc );
    if( r!=MPN_OK ) {
      fprintf( stderr, "error opening news file (free_message_list)\n" );
    }
    docinfo = mpn_news_get_docinfo( newsdoc );

    for( i=0; i<ml->len; i++ ) {

      // set xmms-played
      if( r==MPN_OK ) {
	
	for( j=0; j<mpn_news_get_len( newsdoc ); j++ ) {
	  item = mpn_news_get_item( newsdoc, j );
	  sig = mpn_newsitem_get_sig( item );
	  if( strcmp( ml->list[i]->id, mpn_sig_get_id( sig ) ) == 0 ){
	    app = mpn_newsitem_get_app( item );
	    appblock = mpn_app_get_block( app, "xmms-matrixnews" );
	    if( !appblock ) {
	      appblock = mpn_app_add_block( app, "xmms-matrixnews" );
	    }
	    timestamp = time( NULL );
	    mpn_app_block_set_long( docinfo, appblock, "played_sec", timestamp );
	    mpn_app_block_set_string( docinfo, appblock, "played_txt", ctime( &timestamp ));
	  }
	}
      }
    }
    if( r == MPN_OK ) mpn_news_close( newsdoc );
  }
}

void sort_message_list( message_list_ptr ml )
{
  int i, j;
  message_t_ptr swap;

  if( ml->len>1 ) {
    for( i=ml->len-1; i>=0; i-- ) {
      for( j=0; j<i; j++ ) {
	
	if( ml->list[j]->weight < ml->list[j+1]->weight ) {        
	  swap = ml->list[j];
	  ml->list[j] = ml->list[j+1];
	  ml->list[j+1] = swap;
	}

      }
    }
  }
}

int make_new_message_list( mpn_app_block_ptr nblock, message_list_ptr* ml_ret )
{
  mpn_news_ptr newsdoc, filtered;
  mpn_newsitem_ptr item, filtereditem;
  mpn_sig_ptr sig;
  mpn_file_ptr file;
  mpn_date_ptr expire;
  mpn_net_ptr net;
  mpn_app_ptr app;
  mpn_app_block_ptr appblock;
  mpn_docinfo_ptr docinfo;
  mpn_filter_ptr filter;
  message_list_ptr ml;
  int r, i, j;
  int blockdur, totaldur, newlen;
  char fn[MPN_STRING_LEN];
  char prolog[MPN_STRING_LEN];
  char epilog[MPN_STRING_LEN];
  char separator[MPN_STRING_LEN];
  char cmd[MPN_STRING_LEN];

  r = mpn_news_open_private( MPN_OPEN_READONLY, &newsdoc );
  if( r!=MPN_OK ) {
    fprintf( stderr, "error opening news file\n" );
    return MPN_PERM_ERR;
  }
  docinfo = mpn_news_get_docinfo( newsdoc );

  // Exec system command
  strcpy( cmd, "" );
  mpn_app_block_get_string( docinfo, nblock, "pre_exec", cmd );
  if( strcmp( cmd, "" ) != 0 ) system( cmd );

  r = mpn_app_block_get_filter( docinfo, nblock, "filter", &filter );
  if( r!=MPN_OK ) filter = NULL;

  ml = calloc( sizeof( message_list ), 1 );

  for( i=0; i<mpn_news_get_len( newsdoc ); i++ ) {
    item = mpn_news_get_item( newsdoc, i );

    // reporter?
    app = mpn_newsitem_get_app( item );
    appblock = mpn_app_get_block( app, "reporter" );
    if( !appblock ) {
      continue;
    }

    // !xmms-played
    appblock = mpn_app_get_block( app, "xmms-matrixnews" );
    if( appblock ) {
      if( mpn_app_block_get_long( docinfo, appblock, "played_sec" )>0 ) {
	continue;
      }
    }
    
    // !expired
    sig = mpn_newsitem_get_sig( item );
    expire = mpn_sig_get_expire( sig );
    if( expire ) {
      if( time( NULL )>mpn_date_get_sec( expire ) ) {
	continue;
      }
    }

    // filetype==audio/ogg
    file = mpn_sig_get_file( sig );
    if( !file ) continue;
    if( strcmp( "audio/ogg", mpn_file_get_type( file ) )!=0 ) {
      continue;
    }

    // !corrupt
    net = mpn_newsitem_get_net( item );
    if( mpn_net_get_corrupt( net ) ) {
      continue;
    }

    // filter
    if( filter ) {

      r = mpn_newsitem_filter( item, filter, &filtered, 0 );
      if( r!=MPN_OK ) {
	fprintf( stderr, "error in filter!\n" );
	continue;
      }

      for( j=0; j<mpn_news_get_len( filtered ); j++ ) {
	filtereditem = mpn_news_get_item( filtered, j );
	add_tail_message_list( ml, new_message( docinfo, filtereditem ) );
      }
      mpn_news_close( filtered );

    } else {

      add_tail_message_list( ml, new_message( docinfo, item ) );

    }
  }

  if( filter )
    mpn_filter_free( filter );

  if( ml->len==0 ) {
    free_message_list( ml );
    mpn_news_close( newsdoc );
    return MPN_PERM_ERR;
  }

  // sort list 
  sort_message_list( ml );

  // Cut the tail with duration
  blockdur = mpn_app_block_get_int( docinfo, nblock, "duration" )*60;
  if( blockdur==0 ) blockdur = 15*60;
  totaldur=0; 
  newlen=ml->len;
  for( i=0; i<ml->len; i++ ) {
    if( totaldur<blockdur ) {
      totaldur += ml->list[i]->duration;
    } else {
      if( i<newlen ) newlen=i;
      free( ml->list[i] );
    }
  }

  printf( "newsblock has a duration of %d/%ds, playing %d/%d available messages\n",
	  totaldur, blockdur, newlen, ml->len );
  ml->len = newlen;

  // Fetch messages
  for( i=0; i<ml->len; i++ ) {
    r = fetch_message( ml->list[i], newsdoc );
    if( r!=MPN_OK ) strcpy( ml->list[i]->filename, "" );
  }

  // in case some messages couldnt be loaded 
  j=0;
  for( i=0; i<ml->len; i++ ) {
    if( !(ml->list[i]->filename[0]) ) {
      del_pos_message_list( ml, i );
      i--;
      j++;
    }
  }
  if( j>0 ) {
    printf( "%d messages could not be downloaded\n", j );
  }

  if( ml->len==0 ) {
    free_message_list( ml );
    mpn_news_close( newsdoc );
    return MPN_PERM_ERR;
  }

  // Add prolog and epilog and separators
  mpn_app_block_get_string( docinfo, nblock, "separator", fn );
  if( fn[0] ) {
    sprintf( separator, "%s/%s", mpn_get_workdir(), fn );
    for( i=1; i<ml->len; i+=2 ) {
      add_pos_message_list( ml, new_ogg_jingle_message( separator ), i ); 
    }
  }
  mpn_app_block_get_string( docinfo, nblock, "prolog", fn );
  if( fn[0] ) {
    sprintf( prolog, "%s/%s", mpn_get_workdir(), fn );
    add_pos_message_list( ml, new_ogg_jingle_message( prolog ), 0 );
  }
  mpn_app_block_get_string( docinfo, nblock, "epilog", fn );
  if( fn[0] ) {
    sprintf( epilog, "%s/%s", mpn_get_workdir(), fn );
    add_tail_message_list( ml, new_ogg_jingle_message( epilog ) );
  }

  ml->pos = 0;
  mpn_news_close( newsdoc );
  *ml_ret = ml;
  return MPN_OK;
}

int check_scheduler( int freq, long now, long last, int offset )
{
  struct tm now_tm;
  struct tm last_tm;
  int test, i, lastt, nowt;

  memcpy( &now_tm, localtime( &now ), sizeof(struct tm) );
  memcpy( &last_tm, localtime( &last ), sizeof(struct tm) );

  for( i=0; i<freq; i++ ) {

    test = i*60/freq + offset;
    if( test<0 ) test+=60;
    if( test>59 ) test-=60;
    
    nowt = now_tm.tm_min;
    lastt = last_tm.tm_min;
    if( nowt<lastt ) lastt-=60;
    
    //printf( "check_scheduler: lastt=%d, nowt=%d, test=%d\n", 
    //    lastt, nowt, test );

    if( (test>lastt) && (test<=nowt) ) {
      return 1;
    }

  }
  return 0;
}
 
int make_new_newsblock( message_list_ptr* ml_ret )
{
  mpn_prefs_ptr prefsdoc;
  mpn_app_ptr app;
  mpn_app_block_ptr xblock, nblock;
  mpn_docinfo_ptr docinfo;
  message_list_ptr ml=NULL;
  int r, i, count;
  int freq, offset, preload;
  long now, ft1, ft2;

  r = mpn_prefs_open_private( MPN_OPEN_READONLY, &prefsdoc );
  if( r!=MPN_OK ) {
    fprintf( stderr, "make_new_newsblock: error opening prefs\n" );
    return MPN_PERM_ERR;
  }

  app = mpn_prefs_get_app( prefsdoc );
  if( app ) {
    xblock = mpn_app_get_block( app, "xmms-matrixnews" );
    if( xblock ) {
      docinfo = mpn_prefs_get_docinfo( prefsdoc );
      count = mpn_app_block_get_subblock_count( docinfo, xblock, "newsblock");
      for( i=0; (i<count)&&(i<MAX_NEWSBLOCKS)&&(!ml); i++ )
	  {

		r = mpn_app_block_get_subblock( docinfo, xblock, "newsblock", i,
					&nblock );
		if( r!=MPN_OK ) {
			fprintf( stderr, "make_new_newsblock: error in prefs/newsblock\n" );
			continue;
		}

		now = time( NULL );

		if( i>=newsblocks.len ) {
			newsblocks.len = i+1;
			newsblocks.lastcheck[i] = now;
		}

		freq = mpn_app_block_get_int( docinfo, nblock, "frequency" );
		if( freq==0 ) {
			r = make_new_message_list( nblock, &ml );
			if( r!=MPN_OK ) ml = NULL;
		else {
			printf( "prepared instant newsblock\n" );
			ml->d_time = 0;
		}

	} else {
	  offset = mpn_app_block_get_int( docinfo, nblock, "offset" );
	  preload = mpn_app_block_get_int( docinfo, nblock, "preload" );
	  if( check_scheduler( freq, now, newsblocks.lastcheck[i], 
			       offset-preload ) ) {
	    
	    ft1 = time( NULL );
	    r = make_new_message_list( nblock, &ml );
	    if( r!=MPN_OK ) ml = NULL;
	    else {
	      printf( "prepared scheduled newsblock (frequency=%d, offset=%d, preload=%d)\n", freq, offset, preload );

	      ft2 = time( NULL );
	      ml->d_time = preload*60 - (ft2-ft1);
	    }
	  }
	}

	newsblocks.lastcheck[i] = now;
	mpn_app_block_free( nblock );
      }
    } else {
      printf( "you need app/xmms-matrixnews/newsblock in your prefs file!\n" );
    }
  } else {
    printf( "you need app/xmms-matrixnews/newsblock in your prefs file!\n" );
  }

  mpn_prefs_close( prefsdoc );

  if( ml ) {
    *ml_ret = ml;
    return MPN_OK;
  } else {
    return MPN_PERM_ERR;
  }
}

void* fetch_thread( void* args )
{
  int r;

  while( 1 ) {

    if( playing && !newsing) {
      playing = 0;

      printf( "checking for news\n" );
      r = make_new_newsblock( &play_list );
      if( r==MPN_OK ) {
	printf( "waiting %d seconds until playing\n", play_list->d_time );
	mfl_sleep( (play_list->d_time) * 1000);
	printf( "playing newsblock now\n" );
	newsing = 1;
      }
    }

    mfl_sleep( 10000 );
  }
  return NULL;
}



